import re

class PhraseTemplate:
    def __init__(self, ID=None, TempStrArr = None, TempNLPArr = None, TypeArr = None, Inclinations=None, Values=None, ValuesCZ=None, MatchPattern = None):
        self.ID = ID
        self.TempStr = TempStrArr
        self.TempToNLP = TempNLPArr
        self.Type = TypeArr
        self.Inclinations = Inclinations

        self.tokens_specs = [  # Regex collection for signals  that can occur in conditions
            ("SinO", r"(?P<subject>\w+)in(?P<object>\w+)",r"(?P<subject>\w+)in(?P<object>\w+)",[2,3]),
            ("SatO", r"(?P<subject>\w+)at(?P<object>\w+)",r"(?P<subject>\w+)in(?P<object>\w+)",[2,3]),
            ("SonO", r"(?P<subject>\w+)on(?P<object>\w+)",r"(?P<subject>\w+)in(?P<object>\w+)",[2,3]),
            ("SProgramNum", r"robot(?P<robotName>\w+)programnumber",r"(?P<subject>\w+)in(?P<object>\w+)",[2,3]),
            ("ProgramStarted", r"programstarted",r"(?P<subject>\w+)in(?P<object>\w+)",[2,3]),
            ("ProgramEnded", r"programended",r"(?P<subject>\w+)in(?P<object>\w+)",[2,3])]

        def matchPhrase(self, phrase):
            # totalPhr = []
            # totalTemp = []
            # re.compile(r"robot(?P<robotName>\w+)programnumber", re.IGNORECASE)
            for phr in phrase.values():
                totalPhr += phr
            for str,temp in self.tokens_specs:
                totalTemp += temp
            print("Is {} matching {}?".format(totalPhr, totalTemp))  # (phrase, self.TempStr))
            match = self.TempStr.search(phr)

            # if set(self.MatchPattern) == common:
            #     match = True
            # else:
            #     math = False
            return match





if __name__ == "__main__":

    # phrases are list of SignalPhrase objects
    # mytemplates are list of PhraseTemplate objects obtained from
    mytemplates = PhraseTemplateFactory.make_phrase_dict()
    for phrase in phrases.values():
        phrase.split_phrase()

        PhraseTemp = PhraseTemplate()
        for template in mytemplates.values():
            matching = template.matchPhrase(phrase.splitphrase)
            print(matching)
            if matching:
                phrase.phrasetemplate = template
                # phrase.joinphrase = phrase
                break


from gherkin.parser import Parser
from gherkin.pickles.compiler import compile

from pprint import pprint as pp
import gherkan.utils.constants as c
import os

# requires 'gherkin-official' package
parser = Parser()
input_file = os.path.join(c.DATA_DIR, "input", "test_R1NormalSequence_signals.feature")

with open(input_file, "rt", encoding="utf-8") as file:
    lines = file.read()

    gherkin_document = parser.parse(lines)
    pickles = compile(gherkin_document)

    pp(pickles)


import io
import os

from gherkan.utils import logging_types
import logging

# Imports the Google Cloud client library
from google.cloud import speech
from google.cloud.speech import enums
from google.cloud.speech import types

#if the credentials are inside gherkan package in speech_recognition folder
#os.environ["GOOGLE_APPLICATION_CREDENTIALS"]=os.path.join(os.path.dirname(os.path.realpath(__file__)),
#                                                          "google_credentials.json")
#if the credentials for google api are in the home folder
os.environ["GOOGLE_APPLICATION_CREDENTIALS"]=os.path.join(os.path.expanduser("~/"),"google_credentials.json")                                                          

def transcribe(filename, language_code):
    # Instantiates a client
    try:
        client = speech.SpeechClient()
    except:
        err = "Could not instantiate Google Speech API. Check the path to credentials.",
        logging.error(err, extra={
            "type": logging_types.W_GENERAL_ERROR,
            "phrase": os.environ["GOOGLE_APPLICATION_CREDENTIALS"]})

        raise ConnectionRefusedError(err)

    # Loads the audio into memory
    with io.open(filename, 'rb') as audio_file:
        content = audio_file.read()
        audio = types.RecognitionAudio(content=content)

    config = types.RecognitionConfig(
        encoding=enums.RecognitionConfig.AudioEncoding.LINEAR16,
        language_code=language_code)

    # Detects speech in the audio file
    response = client.recognize(config, audio)
    transcript = " ".join([result.alternatives[0].transcript for result in response.results])

    return transcript


import requests
import unittest
import json


class TestActions(unittest.TestCase):

    def setUp(self):
        self.maxDiff = None
        self.url = f"http://{self.__class__.host}:{self.__class__.port}/{{}}"
        self.testJSON = json.dumps({
            "r1": {
                "1": "zvednout všechny červené kostky",
                "2": "zvednout všechny zelené kostky",
                "3": "zvednout všechny žluté kostky"
            },
            "r2": {
                "1": "dát kostku 1 na podnos",
                "2": "dát kostku 2 na podnos",
                "3": "dát kostku 3 na podnos"
            },
            "r3": {
                "1": "sestavit produkt",
                "2": "dát produkt do krabice"
            }
        })

    @unittest.skip("JSON formatting is buggy.")
    def test_get_and_post_actions(self):
        # originalResponse = requests.get(self.url.format("actions"))
        # print(f"\nCurrent Robot Programs:\n{originalResponse.content.decode()}")
        
        # postResponse = requests.post(self.url.format("actions"), data={"data": self.testJSON}).json()
        # self.assertDictContainsSubset({"OK": True}, postResponse)

        # response = requests.get(self.url.format("actions"))
        # self.assertMultiLineEqual(self.testJSON, response.content.decode())

        # postResponse = requests.post(self.url.format("actions"), data={"data": originalResponse.content}).json()
        # self.assertDictContainsSubset({"OK": True}, postResponse)
        pass




from __future__ import absolute_import

import sys
import unittest
from unittest import TestCase

from lark.tree import Tree

from lark.tools import standalone

try:
    from StringIO import StringIO
except ImportError:
    from io import StringIO

class TestStandalone(TestCase):
    def setUp(self):
        pass

    def _create_standalone(self, grammar):
        code_buf = StringIO()
        temp = sys.stdout
        sys.stdout = code_buf
        standalone.main(StringIO(grammar), 'start')
        sys.stdout = temp
        code = code_buf.getvalue()

        context = {}
        exec(code, context)
        return context

    def test_simple(self):
        grammar = """
            start: NUMBER WORD

            %import common.NUMBER
            %import common.WORD
            %import common.WS
            %ignore WS

        """

        context = self._create_standalone(grammar)

        _Lark = context['Lark_StandAlone']
        l = _Lark()
        x = l.parse('12 elephants')
        self.assertEqual(x.children, ['12', 'elephants'])

    def test_contextual(self):
        grammar = """
        start: a b
        a: "A" "B"
        b: "AB"
        """

        context = self._create_standalone(grammar)

        _Lark = context['Lark_StandAlone']
        l = _Lark()
        x = l.parse('ABAB')

        class T(context['Transformer']):
            def a(self, items):
                return 'a'
            def b(self, items):
                return 'b'
            start = list

        x = T().transform(x)
        self.assertEqual(x, ['a', 'b'])


if __name__ == '__main__':
    unittest.main()





import os
import gherkan.utils.constants as c
from gherkan.speech_recognition.paid_speech_to_text import transcribe

from gherkan.flask_api.raw_text_to_signal import nl_to_signal

audio_path = os.path.join(c.DATA_DIR, "audio", "montrac.wav")
language_code = 'cs-CZ'

text_raw = transcribe(audio_path, language_code)

print(f"Transcribed text: {text_raw}")

request = {
    "feature" : "Montrac",
    "feature_desc" : "Test audia",
    "background" : "Jakmile linka je zapnutá",
    "text_raw" : text_raw,
    "language" : c.LANG_CZ
}

base_path = os.path.join(c.DATA_DIR, "output", "raw_out")

nl_to_signal(base_path, request)

nl_file_path = base_path + ".feature"
signal_file_path = base_path + "_signals.feature"

print("\n\n--- NL FILE ---")
with open(nl_file_path, "rt", encoding="utf-8") as f:
    text = f.read()
    print(text)

print("\n\n--- SIGNAL FILE ---")
with open(signal_file_path, "rt", encoding="utf-8") as f:
    text = f.read()
    print(text)


import atexit
import datetime
import importlib
import logging
import logging.handlers
import queue
import os
import re

import yaml
from flask import send_file

from gherkan.containers.NLPModelManager import NLPModelManager
from .raw_text_to_signal import nl_to_signal as nlToSignal
from .signal_to_nl import signal_to_nl as signalToNL
from .signal_to_negated_signal import signal_to_negated_signal as signalToNegatedSignal
from gherkan.processing.SignalMapper import SignalMapper
from gherkan.speech_recognition.paid_speech_to_text import transcribe as speechToText

DEBUG_MODE = False


class API_FSA():
    """
    FSA class for tracking state from REST API. State is written into a file
    so it should be persistent even over long time periods.
    """

    # =====>>> PATH variables <<<===== #
    stateFilePath = 'state.fsa'  # file to store the current state (see S_* variables)
    configFilePath = 'fsa_config.yaml'  # (editable) configuration file

    # This path does not need to be set to the package's root, probably
    rootFolder = importlib.util.find_spec("gherkan").submodule_search_locations[0]  # the package's root folder

    dataFolder = os.path.join(rootFolder, 'data')  # folder with received and processed files (inp/out files)
    audioFolder = os.path.join(dataFolder, 'audio')  # folder with audio files

    robotProgramsPath = os.path.join(rootFolder, 'utils', 'RobotPrograms_{}.json')  # template of the path to the robot programs
    signalMappingsPath = os.path.join(rootFolder, 'utils', 'signal_mappings_{}.json')  # template of the path to the robot programs

    language = 'en'

    # NLP model manager
    nmm = None

    # =====>>> STATE constants <<<===== #
    # State is stored as 2 bytes in the state.fsa binary file
    S_OFF = 0
    S_IDLE = 1
    S_RECEIVED_NL_TEXT = 2
    S_CORRECTING_NL_TEXT = 4
    S_FINISHED_PROCESSING_NL_TEXT = 8
    S_RECEIVED_SIGNAL = 16
    S_FINISHED_PROCESSING_SIGNAL = 32
    S_RECEIVED_AUDIO = 64
    S_CORRECTING_AUDIO = 128
    S_FINISHED_PROCESSING_AUDIO = 256
    S_SIGNAL_MAPPING_EXISTS = 512
    S_ERROR = 1024

    # This is for debugging
    STATES = ["S_IDLE", "S_RECEIVED_NL_TEXT", "S_CORRECTING_NL_TEXT", "S_FINISHED_PROCESSING_NL_TEXT", "S_RECEIVED_SIGNAL", "S_FINISHED_PROCESSING_SIGNAL", "S_RECEIVED_AUDIO", "S_CORRECTING_AUDIO", "S_FINISHED_PROCESSING_AUDIO", "S_SIGNAL_MAPPING_EXISTS", "S_ERROR"]

    # =====>>> PARAMETER name constants <<<===== #
    P_LAST_SIGNAL_FILE = "lastSignalFile"
    P_LAST_NL_TEXT_FILE = "lastNLTextFile"
    P_LAST_AUDIO_FILE = "lastAudioFile"
    P_RESPONSE_MODE = "responseMode"  # response mode for the processed text, see parameter VALUES

    # =====>>> parameter VALUES <<<===== #
    PV_RESPONSE_FILE = "responseModeFile"  # sends back a file
    PV_RESPONSE_JSON = "responseModeJSON"  # sends back a JSON

    # =====>>> OTHER <<<===== #
    basenameExtractorRegex = re.compile(r"(?P<basename>.+?)(?P<rest>(_signals|\.)+.+)")
    IS_SET_UP = False

    # =====>>> Helper STATIC methods <<<===== #
    @staticmethod
    def toBytes(integer):
        # Maybe extend this with a loop but so far 1024 as the limit is ok
        return bytes([(0b11100000000 & integer) >> 8, 0b11111111 & integer])

    @staticmethod
    def fromBytes(byte):
        # Maybe extend this with a loop but so far 1024 as the limit is ok
        return (byte[0] << 8) | byte[1]

    @staticmethod
    def generateDatestring():
        return datetime.datetime.now().strftime('%Y_%b_%d-%H_%M_%S')

    # =====>>> Helper CLASS methods <<<===== #

    @classmethod
    def states2str(cls):
        with open(cls.stateFilePath, 'rb') as statefile:
            state = cls.fromBytes(statefile.read())
        return [cls.STATES[i] for i in range(10) if (2**i) & state]

    @classmethod
    def getState(cls):
        with open(cls.stateFilePath, 'rb') as statefile:
            state = cls.fromBytes(statefile.read())
        logging.debug("FSA> Requested state: {}", state)
        states = cls.states2str()
        logging.debug("FSA> Current states:\n\t{}", states)
        return state

    @classmethod
    def setState(cls, state):
        logging.debug("FSA> Setting state to: {}", state)
        states = [cls.STATES[i] for i in range(10) if (2**i) & state]
        logging.debug("FSA> Current states:\n\t{}", states)
        if isinstance(state, int):
            state = cls.toBytes(state)
        with open(cls.stateFilePath, 'wb') as statefile:
            statefile.write(state)

    @classmethod
    def addState(cls, state):
        """ Adds a state to the FSA
        This method should be preferred instead of the "setState" method
        """
        currentState = cls.getState()
        if state > cls.S_IDLE:
            currentState &= ~cls.S_IDLE
        cls.setState(currentState | state)

    @classmethod
    def removeState(cls, *states):
        """ Removes a state from the FSA
        This method should be preferred instead of the "setState" method.
        Multiple states can be removed at once
        This methods, however, cannot be used to set the state to S_OFF.
        """
        nextState = cls.getState()
        for state in states:
            nextState &= ~state
        if nextState < cls.S_IDLE:
            nextState |= cls.S_IDLE
        cls.setState(nextState)

    @classmethod
    def getConfig(cls, param=None):
        with open(cls.configFilePath, 'r', encoding="utf-8") as cfg:
            config = yaml.full_load(cfg)
            if param is not None:
                try:
                    value = config[param]
                except Exception as e:
                    value = None
        logging.debug("FSA> Requested value of configuration parameter: {} = {}", param, value)
        return value

    @classmethod
    def setConfig(cls, param, value):
        with open(cls.configFilePath, 'r', encoding="utf-8") as cfg:
            config = yaml.full_load(cfg)
        config[param] = value
        with open(cls.configFilePath, 'w', encoding="utf-8") as cfg:
            yaml.dump(config, cfg)
        logging.debug("FSA> Setting configuration parameter: {} = {}", param, value)
        logging.debug("Configuration yaml file:\n\t{}", str(config))

    @classmethod
    def regenerateConfig(cls):
        logging.debug("FSA> Regenerating configuration file")
        with open(cls.configFilePath, 'w', encoding="utf-8") as cfg:
            # TODO: add more defaults to config
            yaml.dump({"flush": False, "responseMode": "responseModeFile"}, cfg)

    # =====>>> API initialization <<<===== #
    @classmethod
    def setup(cls, debugMode=False):
        """This function sets up the the paths and logging

        Parameters
        ----------
        debugMode : bool, optional
            This enforces the logging system to output the logging messages as well (the default is False, which [default_description])
        """

        global DEBUG_MODE
        if not DEBUG_MODE:  # Manually setting DEBUG_MODE to True overrides the setting requested via the API
            DEBUG_MODE = debugMode

        # Check if the root folder is set correctly
        if not os.path.isdir(cls.rootFolder):
            raise Exception("The root folder is set to a non-existent path!\nRoot folder: {}".format(cls.rootFolder))
        if not os.access(cls.rootFolder, os.W_OK):
            raise Exception("Cannot write to the root folder!\nRoot folder: {}".format(cls.rootFolder))
        if not os.access(cls.rootFolder, os.R_OK):
            raise Exception("Cannot read from the root folder!\nRoot folder: {}".format(cls.rootFolder))

        # Check the existence of data folders
        if not os.path.isdir(cls.audioFolder):
            os.makedirs(cls.audioFolder)

        # > LOGGING <
        logging.captureWarnings(True)  # capture warnings

        class QFilter(logging.Filter):

            def filter(self, record):
                return hasattr(record, "type")

        class ConsoleFormatter(logging.Formatter):

            def __init__(self):
                self.default_msec_format = "%s.%03d"
                super().__init__(datefmt="%Y-%m-%d %H:%M:%S.uuu")
                self.oldFormatterRegex = re.compile(r"%\w+:")

            def format(self, record):
                template = "[{time}] *{severity}*: "
                if hasattr(record, "type"):
                    template = template + f"[{record.type}] {{message}}"
                else:
                    template = template + "{message}"

                if record.args:
                    if self.oldFormatterRegex.search(record.msg) is not None:
                        formattedMessage = record.msg % record.args
                    else:
                        formattedMessage = record.msg.format(*record.args)
                else:
                    formattedMessage = record.msg
                record.message = formattedMessage
                string = template.format(severity=record.levelname, time=self.formatTime(record), message=formattedMessage)

                if record.levelno >= logging.ERROR:
                    string += f"\nError occurred in file '{record.filename}' ({record.pathname}) in function '{record.funcName}' on line '{record.lineno}'. See the log file for more details."
                string += ""
                return string

        class NewStyleLR(logging.LogRecord):

            def __init__(self, name, level, pathname, lineno, msg, args, exc_info, func, sinfo):
                super().__init__(name, level, pathname, lineno, msg, args, exc_info, func, sinfo)
                self.oldFormatterRegex = re.compile(r"%\w")

            def getMessage(self):
                msg = str(self.msg)
                if self.args:
                    if self.oldFormatterRegex.search(msg) is not None:
                        msg = msg % self.args
                    else:
                        msg.format(*self.args)
                self.message = msg
                return msg

        logging.setLogRecordFactory(NewStyleLR)

        cls.logQueue = queue.Queue()  # Queue to put the log records into
        queueHandler = logging.handlers.QueueHandler(cls.logQueue)
        queueHandler.addFilter(QFilter())
        queueHandler.setFormatter(ConsoleFormatter())
        queueHandler.setLevel(logging.WARNING)

        consoleHandler = logging.StreamHandler()
        consoleHandler.setLevel(logging.DEBUG if DEBUG_MODE else logging.INFO)
        consoleHandler.setFormatter(ConsoleFormatter())

        fileHandler = logging.FileHandler(filename=os.path.join(cls.dataFolder, "log.txt"), encoding="utf-8")
        fileHandler.setLevel(logging.NOTSET)
        fileHandler.setFormatter(ConsoleFormatter())

        logging.basicConfig(handlers=[queueHandler, fileHandler, consoleHandler])
        rootLogger = logging.getLogger()  # get the root logger
        rootLogger.setLevel(0)  # the logger filters the records even before they get to handlers, so the level shall be set to lowest possible value

        if not os.path.isfile(cls.configFilePath):
            cls.regenerateConfig()

        cls.IS_SET_UP = True

    # =====>>> API initialization <<<===== #
    @classmethod
    def init_fsa(cls, address, port):
        if not cls.IS_SET_UP:
            raise Exception("The API FSA was not setup properly before initialization! Run API_FSA.setup() before init_fsa()!")
        # > Initialization <
        logging.info("The RESTful API to the Gherkan NL Instruction Processing system is starting...")

        logging.info("Host address: {}\nPort number: {}", address, port)
        cls.setConfig("host", address)
        cls.setConfig("port", port)

        logging.info("Checking done, checking state from previous run...")
        # Check for previous sessions
        if os.path.exists(cls.stateFilePath):
            previousState = cls.getState()
            # Check and handle error
            if previousState == cls.S_ERROR:
                if cls.getConfig('flush'):
                    logging.info("Found error state from a previous run.\nError flushing is set to True - flushing the error and reseting the system.")
                    previousState = cls.S_OFF
                    cls.regenerateConfig()
                else:
                    raise Exception("Previous run of the system resulted in an error! Resolve the error and set the fsa_config\nPlease, check the log file for details.")
            if previousState != cls.S_OFF:
                logging.info("Found non-OFF state from previous run, restoring the system with state {}", previousState)
                cls.setState(previousState)
            else:
                # Set state to ready/idle
                logging.info("System ready, state set to IDLE.")
                cls.setState(cls.S_IDLE)
        else:
            logging.info("System ready, state set to IDLE.")
            cls.setState(cls.S_IDLE)

        # Load NLP models (time-consuming operation)
        logging.info("Loading NLP models")

        cls.nmm = NLPModelManager()
        cls.nmm.load_all_models()

        # Register function that will shutdown the FSA when the server is terminated
        atexit.register(cls.finalize)

    @classmethod
    def finalize(cls):
        """Shuts down the FSA when the server (application) is terminated
        """

        logging.info("Shutting down FSA.")
        cls.setState(cls.S_OFF)
        logging.shutdown()

    # =====>>> RESTful API Call handling functions <<<===== #
    @classmethod
    def receiveSignal(cls, text):
        """This function handles the receiving of signal text.
        """

        state = cls.getState()
        if state & (cls.S_RECEIVED_AUDIO | cls.S_RECEIVED_NL_TEXT):
            raise Exception("Different data type already received!")

        if state & cls.S_RECEIVED_SIGNAL:
            signalFileMode = 'a'
            fileName = cls.getConfig(cls.P_LAST_SIGNAL_FILE)
        else:
            cls.addState(cls.S_RECEIVED_SIGNAL)
            signalFileMode = 'w'
            fileName = os.path.join(cls.dataFolder, '_'.join([cls.generateDatestring(), 'signals.feature']))
            cls.setConfig(cls.P_LAST_SIGNAL_FILE, fileName)

        with open(fileName, signalFileMode, encoding="utf-8") as sigFile:
            sigFile.writelines(text)

        cls.removeState(cls.S_SIGNAL_MAPPING_EXISTS)

        # Signal text can be iteratively written into the file
        # When all signals were written, the user can request NL file

    @classmethod
    def requestSignal(cls):
        """This function return a composed signal file derived from NL text, if it was provided
        """
        state = cls.getState()
        if state & cls.S_RECEIVED_SIGNAL:  # Signals were sent and requested back (e.g. to check the state of the file)
            return send_file(cls.getConfig(cls.P_LAST_SIGNAL_FILE))
        elif not (state & cls.S_FINISHED_PROCESSING_NL_TEXT):  # Check if NL text was sent and processed
            raise Exception("NL text was not provided, cannot return signal file!")

        signalFilePath = cls.getConfig(cls.P_LAST_SIGNAL_FILE)

        responseMode = cls.getConfig(cls.P_RESPONSE_MODE)
        if responseMode == cls.PV_RESPONSE_FILE:
            response = send_file(signalFilePath)
        elif responseMode == cls.PV_RESPONSE_JSON:
            with open(signalFilePath, "r", encoding="utf-8") as signalFile:
                signalText = signalFile.read()

            # TODO: Split file into response

            response = {
                "language": "en<OR>cs",
                "background": "background text",
                "description": "This feature is not implemented, yet! Use RESPONSE_MODE: FILE instead.",
                "scenarios": signalText
            }

        cls.removeState(cls.S_FINISHED_PROCESSING_NL_TEXT, cls.S_RECEIVED_NL_TEXT)
        return response

    @classmethod
    def requestRemappedSignal(cls):
        """This function return a remapped signal
        """
        state = cls.getState()
        if not state & (cls.S_RECEIVED_SIGNAL | cls.S_FINISHED_PROCESSING_NL_TEXT):
            Exception("NL text nor signal file was not provided, cannot return signal file!")

        if not state & cls.S_SIGNAL_MAPPING_EXISTS:
            return send_file(cls.getConfig(cls.P_LAST_SIGNAL_FILE))

        signalFilePath = cls.getConfig(cls.P_LAST_SIGNAL_FILE)

        path, filename = os.path.split(signalFilePath)
        remappedSignalFilePath = os.path.join(path, cls.basenameExtractorRegex.sub(r"\g<basename>_remapped\g<rest>", filename))
        
        signalMapper = SignalMapper()

        signalMapper.loadSignal(signalFilePath)  # load the signal file
        signalMapper.loadDictionary(cls.signalMappingsPath.format(cls.language))  # write the signal file
        signalMapper.encode()  # encode the signal with the loaded dictionary
        signalMapper.writeSignal(remappedSignalFilePath)

        response = send_file(remappedSignalFilePath)
        # cls.removeState(cls.S_FINISHED_PROCESSING_NL_TEXT, cls.S_RECEIVED_NL_TEXT)
        return response

    @classmethod
    def requestNegatedSignal(cls):
        """This function return the signal file with all the statements negated.
        """
        state = cls.getState()

        if not (state & cls.S_RECEIVED_SIGNAL):
            raise Exception("Signal file was not provided, cannot return negated signal file!")

        signal_file = cls.getConfig(cls.P_LAST_SIGNAL_FILE)

        outputFile = signalToNegatedSignal(signal_file, cls.nmm)

        cls.removeState(cls.S_RECEIVED_SIGNAL)

        responseMode = cls.getConfig(cls.P_RESPONSE_MODE)
        if responseMode == cls.PV_RESPONSE_FILE:
            response = send_file(outputFile)
        else:
            response = {"Message": "Response mode not implemented yet."}

        return response

    @classmethod
    def requestNLText(cls):
        state = cls.getState()
        if not (state & cls.S_RECEIVED_SIGNAL):
            raise Exception("Signal file was not provided, cannot return NL text!")

        signalFilePath = cls.getConfig(cls.P_LAST_SIGNAL_FILE)

        path, filename = os.path.split(signalFilePath)
        outputFile = os.path.join(path, cls.basenameExtractorRegex.sub(r"\g<basename>.feature", filename))

        logging.debug("Calling SignalToNL; associated file: {}", signalFilePath)
        signalToNL(signalFilePath, cls.nmm)

        cls.removeState(cls.S_RECEIVED_SIGNAL)

        responseMode = cls.getConfig(cls.P_RESPONSE_MODE)
        if responseMode == cls.PV_RESPONSE_FILE:
            response = send_file(outputFile)
        elif responseMode == cls.PV_RESPONSE_JSON:
            with open(outputFile, "r", encoding="utf-8") as nlTextFile:
                nlText = nlTextFile.read()

            response = {
                "language": "en<OR>cz",
                "description": "This feature is not implemented, yet! Use RESPONSE_MODE: FILE instead.",
                "background": "background/context text",
                "scenarios": nlText
            }
        else:
            response = {"Message": "Request mode is set to an incorrect value!"}
        return response

    @classmethod
    def receiveNLText(cls, data):
        state = cls.getState()
        if state & (cls.S_RECEIVED_SIGNAL | cls.S_RECEIVED_AUDIO):
            raise Exception("Different data type already received!")

        cls.addState(cls.S_RECEIVED_NL_TEXT)

        basepath = os.path.join(cls.dataFolder, cls.generateDatestring())
        nlFilePath = f"{basepath}.feature"
        signalFilePath = f"{basepath}_signals.feature"

        logging.debug("Calling nlToSignal; associated file: {}\nData: {}", basepath, data)
        cls.getLogs()  # this will empty any previous log messages, just in case
        nlToSignal(basepath, data, cls.nmm)
        logs = cls.getLogs()  # get log records generated while processing the NL text

        response = {
            "info": "Processing done",
            "lines": [],
            "error_lines": [],
            "error_hints": [],
            "errors": False
        }

        if len(logs) > 0:  # There were some problems while processing the text
            cls.removeState(cls.S_RECEIVED_NL_TEXT)
            response["info"] = "Some problems occurred while processing the NL text. Please correct the indicated problems and resubmit the text."
            response["errors"] = True
            for log in logs:
                response["error_lines"].append(log.message)
        else:
            cls.setConfig(cls.P_LAST_NL_TEXT_FILE, nlFilePath)
            cls.setConfig(cls.P_LAST_SIGNAL_FILE, signalFilePath)
            cls.addState(cls.S_FINISHED_PROCESSING_NL_TEXT)

        cls.removeState(cls.S_SIGNAL_MAPPING_EXISTS)

        return response

    @classmethod
    def receiveAudio(cls, audioPath, language):
        state = cls.getState()
        if state & (cls.S_RECEIVED_SIGNAL | cls.S_RECEIVED_NL_TEXT):
            raise Exception("Different data type already received!")

        if language == "en":
            lang = "en-US"
        elif language == "cs":
            lang = "cs-CZ"
        transcriptPath = '.'.join([audioPath[:audioPath.find(".")], "txt"])

        logging.debug("Calling speechToText; audio file: {}\nlanguage: {}", audioPath, lang)
        try:
            transcript = speechToText(audioPath, lang)
        except Exception as e:
            raise e

        logging.info(f"Audio transcript: {transcript}")
        logging.info(f"Saving transcript to {transcriptPath}")

        with open(transcriptPath, "w", encoding="utf-8") as tf:
            tf.writelines(transcript)

        return transcript

    @classmethod
    def requestRobotPrograms(cls, language):
        if not language:
            language = cls.language

        with open(cls.robotProgramsPath.format(language), 'r', encoding="utf-8") as robotProgramFile:
            robotPrograms = robotProgramFile.read()

        return robotPrograms

    @classmethod
    def receiveRobotPrograms(cls, programs, language):
        if not language:
            language = cls.language
        programs.save(cls.robotProgramsPath.format(language))
        programs.close()

    @classmethod
    def requestSignalMapping(cls, language):
        state = cls.getState()
        if not (state & (cls.S_RECEIVED_SIGNAL | cls.S_FINISHED_PROCESSING_NL_TEXT)):
            raise Exception("There is no signal file to extract the mapping from!")

        if not language:
            language = cls.language

        signalFilePath = cls.getConfig(cls.P_LAST_SIGNAL_FILE)  # get the last signal file

        signalMapper = SignalMapper()

        signalMapper.loadSignal(signalFilePath)  # load the signal file
        signalMapper.analyze()  # analyze it (get the signals)

        signalMapper.writeDictionary(cls.signalMappingsPath.format(language))  # write the signal file

        cls.addState(cls.S_SIGNAL_MAPPING_EXISTS)

        with open(cls.signalMappingsPath.format(language), 'r', encoding="utf-8") as signalMappingsFile:
            signalMappings = signalMappingsFile.read()  # the file is written and reread - not efficient but easier to code...

        return signalMappings

    @classmethod
    def receiveSignalMapping(cls, mapping, language):
        state = cls.getState()
        if not (state & cls.S_SIGNAL_MAPPING_EXISTS):
            raise Exception("Signal mapping was not created! First, request the signal mapping file then adjustment to the mapping can be made.")

        if not language:
            language = cls.language

        mapping.save(cls.signalMappingsPath.format(cls.language))
        mapping.close()

    # =====>>> RESTful API Helper functions <<<===== #
    @classmethod
    def getLogs(cls):
        logs = []
        while not cls.logQueue.empty():
            try:
                item = cls.logQueue.get(False)
            except queue.Empty:
                logging.exception("The logging queue was emptied for some reason (even though while not empty is used).")
            else:
                logs.append(item)
        return logs

    @classmethod
    def canRequestSignal(cls):
        """Checks whether it is possible to request a signal file.
        """
        state = cls.getState()
        return state & (cls.S_FINISHED_PROCESSING_NL_TEXT | cls.S_RECEIVED_SIGNAL)

    @classmethod
    def canRequestNegatedSignal(cls):
        """Checks whether it is possible to request a negated signal file.
        """
        state = cls.getState()
        return state & cls.S_RECEIVED_SIGNAL

    @classmethod
    def canRequestNLText(cls):
        """Checks whether it is possible to request a NL text file.
        """
        state = cls.getState()
        return state & cls.S_RECEIVED_SIGNAL



from .signal_scenario import SignalScenario


class SignalBatch():
    """
    Container class for batch of test scenarios encoded into signals.

    ...

    Attributes
    ----------

    language : str
        language abbreviation (e.g. "cs" or "en") to indicate the language of the text

    name : str
        name of the batch (can be the same as filename)

    filename : str
        name of the original filename from which was the batch loaded (if available)

    desc : str
        description and/or precondition of the batch

    context : str
        context for the "while" conditions of the included text scenarios (i.e. a condition that will apply to all scenarios)

    scenarios : list
        list of SignalScenario objects that are part of this batch
    """

    def __init__(self, language=None, name="", filename="", desc="", context=""):
        self.language = language
        self.name = name
        self.filename = filename
        self.desc = desc
        self.context = context
        self.scenarios = []

    @property
    def filename(self):
        return self.__filename

    @filename.setter
    def filename(self, filename):
        self.__filename = filename
        if not self.name:
            self.name = filename

    def addScenario(self, scenario: SignalScenario):
        self.scenarios.append(scenario)



# coding: utf-8
from gherkan.processing import Temp2NL

fill_action = Temp2NL.FillActions()
Name = "PNStart"
VarsVals = ['robotR1', 2]
Type = "equality"
Values = "1"
TempStr="start"

TempData = ["PNStart", ['robotR1', 2], "equality", "1", "start"]
print(TempData)
sentence_core_part = fill_action.ParseMainAction(TempData)
print(sentence_core_part)

TempData = ["sDest", ['shuttleXY', 2], "equality", "1", "destination is"]
print(TempData)
sentence_core_part = fill_action.ParseMainAction(TempData)
print(sentence_core_part)

TempData = ["sState", ['storageXY'], "bool", "1", "free"]
print(TempData)
sentence_core_part = fill_action.ParseMainAction(TempData)
print(sentence_core_part)

TempData = ["sAct", ['shuttleX'], "bool", "1", "move"]
print(TempData)
sentence_core_part = fill_action.ParseMainAction(TempData)
print(sentence_core_part)

TempData = ["sAct", ['shuttleX', "stationZ"], "bool", "1", "move to"]
print(TempData)
sentence_core_part = fill_action.ParseMainAction(TempData)
print(sentence_core_part)

# TempData = ["sCount", ['storageX', "5"], "equality", "5", "count is"]
# print(TempData)
# sentence_core_part = fill_action.ParseMainAction(TempData)
# print(sentence_core_part)





# -*- coding: utf-8 -*-
from gherkan.utils import constants as c


class StatementTreeNode():
    def __init__(self, kind, parent):
        self.kind = kind
        self.parent = parent
        self.negated = False


    def negate(self):
        self.negated = not self.negated


    def isNegated(self):
        return self.negated


class StatementTreeBinaryOperatorNode(StatementTreeNode):
    def __init__(self, kind, parent=None):
        super().__init__(kind, parent)
        self.lchild = None
        self.rchild = None

    def __str__(self):
        return 'Operator(type: {})'.format(self.kind)

    def negate(self):
        if self.kind == c.AND:
            self.kind = c.OR
        elif self.kind == c.OR:
            self.kind = c.AND
        else:
            raise NotImplementedError("Cannot negate {}".format(self.kind))


class StatementTreeOperandNode(StatementTreeNode):
    def __init__(self, kind, parent=None):
        super().__init__(kind, parent)
        self.data = StatementTreeNodeData()

    def __str__(self):
        return 'Operand(type: {}, variable: {}, value: {}, variableNLP: {})'.format(
            self.kind, self.data.variable, self.data.value, self.data.variableNL)


class StatementTreeMergedNode(StatementTreeNode):
    def __init__(self, kind="MERGED_NODE", parent=None):
        super().__init__(kind, parent)
        self.subnodes = []
        self.data = StatementTreeNodeData()

    def __str__(self):
        return 'MergedNode(contains {} nodes)'.format(len(self.subnodes))

class StatementTreeNodeData():
    def __init__(self):
        self.string = None
        self.variable = None
        self.value = None
        self.variableNL = None
        self.valueId = None


import re
from functools import partial

from .utils import get_regexp_width
from .parsers.grammar_analysis import GrammarAnalyzer
from .lexer import TraditionalLexer, ContextualLexer, Lexer, Token
from .parsers import lalr_parser, earley, xearley, cyk
from .tree import Tree

class WithLexer:
    lexer = None
    parser = None
    lexer_conf = None

    def init_traditional_lexer(self, lexer_conf):
        self.lexer_conf = lexer_conf
        self.lexer = TraditionalLexer(lexer_conf.tokens, ignore=lexer_conf.ignore, user_callbacks=lexer_conf.callbacks)

    def init_contextual_lexer(self, lexer_conf):
        self.lexer_conf = lexer_conf
        states = {idx:list(t.keys()) for idx, t in self.parser._parse_table.states.items()}
        always_accept = lexer_conf.postlex.always_accept if lexer_conf.postlex else ()
        self.lexer = ContextualLexer(lexer_conf.tokens, states,
                                     ignore=lexer_conf.ignore,
                                     always_accept=always_accept,
                                     user_callbacks=lexer_conf.callbacks)

    def lex(self, text):
        stream = self.lexer.lex(text)
        if self.lexer_conf.postlex:
            return self.lexer_conf.postlex.process(stream)
        return stream

    def parse(self, text):
        token_stream = self.lex(text)
        sps = self.lexer.set_parser_state
        return self.parser.parse(token_stream, *[sps] if sps is not NotImplemented else [])

class LALR_TraditionalLexer(WithLexer):
    def __init__(self, lexer_conf, parser_conf, options=None):
        debug = options.debug if options else False
        self.parser = lalr_parser.Parser(parser_conf, debug=debug)
        self.init_traditional_lexer(lexer_conf)

class LALR_ContextualLexer(WithLexer):
    def __init__(self, lexer_conf, parser_conf, options=None):
        debug = options.debug if options else False
        self.parser = lalr_parser.Parser(parser_conf, debug=debug)
        self.init_contextual_lexer(lexer_conf)

class LALR_CustomLexer(WithLexer):
    def __init__(self, lexer_cls, lexer_conf, parser_conf, options=None):
        self.parser = lalr_parser.Parser(parser_conf)
        self.lexer_conf = lexer_conf
        self.lexer = lexer_cls(lexer_conf)

def tokenize_text(text):
    line = 1
    col_start_pos = 0
    for i, ch in enumerate(text):
        if '\n' in ch:
            line += ch.count('\n')
            col_start_pos = i + ch.rindex('\n')
        yield Token('CHAR', ch, line=line, column=i - col_start_pos)

class Earley(WithLexer):
    def __init__(self, lexer_conf, parser_conf, options=None):
        self.init_traditional_lexer(lexer_conf)

        resolve_ambiguity = options.ambiguity == 'resolve'
        self.parser = earley.Parser(parser_conf, self.match, resolve_ambiguity=resolve_ambiguity)

    def match(self, term, token):
        return term.name == token.type


class XEarley:
    def __init__(self, lexer_conf, parser_conf, options=None, **kw):
        self.token_by_name = {t.name:t for t in lexer_conf.tokens}

        self._prepare_match(lexer_conf)
        resolve_ambiguity = options.ambiguity == 'resolve'
        self.parser = xearley.Parser(parser_conf,
                                    self.match,
                                    ignore=lexer_conf.ignore,
                                    resolve_ambiguity=resolve_ambiguity,
                                    **kw
                                    )

    def match(self, term, text, index=0):
        return self.regexps[term.name].match(text, index)

    def _prepare_match(self, lexer_conf):
        self.regexps = {}
        for t in lexer_conf.tokens:
            regexp = t.pattern.to_regexp()
            try:
                width = get_regexp_width(regexp)[0]
            except ValueError:
                raise ValueError("Bad regexp in token %s: %s" % (t.name, regexp))
            else:
                if width == 0:
                    raise ValueError("Dynamic Earley doesn't allow zero-width regexps", t)

            self.regexps[t.name] = re.compile(regexp)

    def parse(self, text):
        return self.parser.parse(text)

class XEarley_CompleteLex(XEarley):
    def __init__(self, *args, **kw):
        XEarley.__init__(self, *args, complete_lex=True, **kw)



class CYK(WithLexer):

    def __init__(self, lexer_conf, parser_conf, options=None):
        self.init_traditional_lexer(lexer_conf)

        self._analysis = GrammarAnalyzer(parser_conf)
        self._parser = cyk.Parser(parser_conf.rules, parser_conf.start)

        self._postprocess = {}
        for rule in parser_conf.rules:
            a = rule.alias
            self._postprocess[a] = a if callable(a) else (a and getattr(parser_conf.callback, a))

    def parse(self, text):
        tokens = list(self.lex(text))
        parse = self._parser.parse(tokens)
        parse = self._transform(parse)
        return parse

    def _transform(self, tree):
        subtrees = list(tree.iter_subtrees())
        for subtree in subtrees:
            subtree.children = [self._apply_callback(c) if isinstance(c, Tree) else c for c in subtree.children]

        return self._apply_callback(tree)

    def _apply_callback(self, tree):
        children = tree.children
        callback = self._postprocess[tree.rule.alias]
        assert callback, tree.rule.alias
        r = callback(children)
        return r


def get_frontend(parser, lexer):
    if parser=='lalr':
        if lexer is None:
            raise ValueError('The LALR parser requires use of a lexer')
        elif lexer == 'standard':
            return LALR_TraditionalLexer
        elif lexer == 'contextual':
            return LALR_ContextualLexer
        elif issubclass(lexer, Lexer):
            return partial(LALR_CustomLexer, lexer)
        else:
            raise ValueError('Unknown lexer: %s' % lexer)
    elif parser=='earley':
        if lexer=='standard':
            return Earley
        elif lexer=='dynamic':
            return XEarley
        elif lexer=='dynamic_complete':
            return XEarley_CompleteLex
        elif lexer=='contextual':
            raise ValueError('The Earley parser does not support the contextual parser')
        else:
            raise ValueError('Unknown lexer: %s' % lexer)
    elif parser == 'cyk':
        if lexer == 'standard':
            return CYK
        else:
            raise ValueError('CYK parser requires using standard parser.')
    else:
        raise ValueError('Unknown parser: %s' % parser)






from flask import jsonify, request
from flask_restful import Resource, reqparse, abort
from .. import API_FSA
import logging

class Actions(Resource):
    parser = reqparse.RequestParser()

    def get(self, language=''):
        return API_FSA.requestRobotPrograms(language)

    def post(self, language=''):
        try:
            for _, value in request.files.items():
                if not language:
                    language = request.form["language"]
                API_FSA.receiveRobotPrograms(value, language)
            return {"OK": True}
        except Exception as error:
            errorMessage = f"Unable to receive the robot programs! {error}"
            logging.exception(errorMessage)
            abort(400, message=errorMessage)



# coding: utf-8
"""
Class FillActions turns the filled object SignalPhrase into a natural language phrase, which is then returned
as the filled field niceStr in the SignalPhrase.
"""
import json, os, re
import yaml
from pattern.en import PARTICIPLE, conjugate
import gherkan.utils.constants as c
import gherkan.utils.gherkin_keywords as g
from gherkan.containers.NLPModelManager import NLPModelManager
from gherkan.utils.word_morpho import Finetuner
from gherkan.utils import logging_types
import logging

CZ_PROGRAM_DICT_DIR = os.path.join(c.GHERKAN_ROOT_DIR, "utils", "RobotPrograms_cs.json")
EN_PROGRAM_DICT_DIR = os.path.join(c.GHERKAN_ROOT_DIR, "utils", "RobotPrograms_en.json")
GENERAL_YAML_DIR = os.path.join(c.GHERKAN_ROOT_DIR, "utils", "en-cz_phrases.yaml")


class FillActions:
    def __init__(self, nmm : NLPModelManager):
        self.finetune = Finetuner(lang="cs", nmm=nmm)
        self.lang = None

    def ParseMainAction(self, data):
        """ Construct basic NL phrase from given template
            Input format of data:
            data = {Name, Vars, Values, TempToNL}
            Vars = list of strings; specific values to fill into template
            Values = list of integers; values for given Types
            TempToNL = string; specific details for given template
        """
        self.data = self.convert_to_dic(data)
        self.lang = self.data["language"]
        varz = self.data["vars"]
        values = self.data["values"]
        phrase = self.data["tempToNL"][0]
        for key in varz:  # separate actor names (shuttleXY > shuttle XY)
            if not varz[key].isupper():
                string = str(varz[key][1:])  # we leave out the first letter in case it is upper
                actor = (re.findall('([A-Z0-9].*)', string))
                if actor:
                    varz[key] = varz[key].replace(actor[0], " {}".format(actor[0]))
                else:
                    pass
        for key, v in values.items():  # try to match the values with .json
            matched = self.MatchAction(varz['subject'], values[key])
            if matched:
                values[key] = matched
                to_replace = re.compile(r"#\(\?P\<{}\>[^\)]+\)".format(key))
                phrase = re.sub(to_replace, values[key], phrase)
            else:
                if isinstance(values[key], bool):
                  phrase = self.ParseBool(phrase, values[key])
                else:
                    to_replace = re.compile(r"\(\?P\<{}\>[^\)]+\)".format(key))
                    phrase = re.sub(to_replace, values[key], phrase)
        for key, value in varz.items():  # match varz with template
            idx = list(varz).index(key)
            if key == "place":
                value = value + "_2"
            value = self.MatchGeneral(value)
            print(value)
            if self.data["inclinations"][idx] != 0:
                value = self.finetune.conjugate(self.lang, value, self.data["inclinations"][idx])
                print(value)
            to_replace = re.compile(r"\(\?P\<{}\>[^\)]+\)".format(key))
            phrase = re.sub(to_replace, value, phrase)
        verb = self.finetune.find_verb(self.lang, phrase)
        if verb:
            verb_conj = self.finetune.conjugate(self.lang, verb[0], 3) # conjugate verb to third person
            phrase = re.sub(verb[0], verb_conj, phrase)
        for word in phrase.split():
            word_new = self.MatchGeneral(word)  # translate words to czech
            phrase = re.sub(r'\b({})\b'.format(word), word_new, phrase)
        print(phrase)
        phrase = self.finetune.strip_extra_spaces(phrase)
        phrase = self.finetune.align_adjectives(self.lang, phrase)

        if phrase:
            data.__dict__["niceStr"] = phrase
        else:
            data.__dict__["niceStr"] = self.data["niceStr"]
            logging.debug("Failed to create NL description", extra={
                "type": logging_types.W_TEMPLATE_NOT_FOUND,
                "phrase": self.data["niceStr"]})


    def ParseBool(self, phrase, value):
        """ Bool 1 is positive, Bool 0 negative (sth does not happen) """
        exceptions = ["be", "can", "will", "must"]
        if value:
            phrase_bool = phrase
        elif not value:
            verb = self.finetune.find_verb(self.lang, phrase)
            if verb:
                if self.lang == "cs":
                    phrase_bool = re.sub(verb[0], "".join(["ne", verb[0]]), phrase)
                elif self.lang == "en":
                    verb_lemma = self.finetune.lemmatize(self.lang, verb[0])
                    if verb_lemma in exceptions:
                        phrase_bool = re.sub(r'\b{}\b'.format(verb[0]),  " ".join([verb_lemma, "not"]), phrase)
                    else:
                        phrase_bool = re.sub(r'\b{}\b'.format(verb[0]), " ".join(["does not", verb_lemma]), phrase)
            else:
                phrase_bool = phrase
        return phrase_bool

    def ParseForce(self, phrase, value):
        if value:
            if self.lang == "en":
                phrase = " ".join(["force", phrase])
            elif self.lang == "cs":
                phrase = " ".join(["Vynuť", phrase])
        elif not value:
            phrase = phrase
        return phrase

    def convert_to_dic(self, lists):
        data = lists.__dict__
        return data

    def strip_extra_spaces(self, text):
        stripped_spaces = re.sub(' +', ' ', text)
        stripped_text = stripped_spaces.strip()
        return stripped_text

    def start_stop(self, phrase, state):
        state = self.finetune.conjugate(self.lang, state.lower(), 3)
        verb = self.finetune.find_verb(self.lang, phrase)
        verb_new = verb[0]
        if verb:
            if self.lang == "cs":
                verb_new = self.finetune.lemmatize(self.lang, verb[0])
            elif self.lang == "en":
                verb_new = conjugate(verb[0], tense=PARTICIPLE, parse=True)
            phrase = re.sub(verb[0], " ".join([state, verb_new]), phrase)
            return phrase
        else:
            return phrase

    def MatchAction(self, actor, prog):
        """ Load robot actions and return sring for given program number """
        actor = actor.lower()
        if self.lang == "en":
            with open(EN_PROGRAM_DICT_DIR, 'r', encoding="utf-8") as stream:
                action_list = json.load(stream)
                if actor in action_list:
                    if str(prog) in action_list[actor]:
                        action = action_list[actor][str(prog)]
                    else:
                        action = "execute program {}".format(prog)
                        logging.warning("Warning: Did not find program {} for robot {} in English", prog, actor,
                        extra={"type": logging_types.W_TEMPLATE_NOT_FOUND})
                else:
                    return False

        elif self.lang == "cs":
            with open(CZ_PROGRAM_DICT_DIR, 'r', encoding="utf-8") as stream:
                action_list = json.load(stream)
                if actor in action_list:
                    if str(prog) in action_list[actor]:
                        action = action_list[actor][str(prog)]
                    else:
                        action = "vykonat program {}".format(prog)
                        logging.warning("Warning: Did not find program {} for robot {} in Czech", prog, actor,
                        extra={"type": logging_types.W_TEMPLATE_NOT_FOUND})
                else:
                    return False
        return action

    def MatchGeneral(self, word):
        """ Return nice string for basic form of a word """
        order = 0
        if "_2" in word:
            order = 1
            word = word.split("_")[0]
        with open(GENERAL_YAML_DIR, 'r', encoding="utf-8") as stream:
            phrase_list = yaml.full_load(stream)
        if self.lang in phrase_list and word.lower() in phrase_list[self.lang]:
            phrase = phrase_list[self.lang][word.lower()]
            if type(phrase) is list:
                phrase = phrase[order]
        else:
            phrase = word
        return phrase





# -*- coding: utf-8 -*-
from flask import jsonify, request, Response
from flask_restful import Resource, reqparse, abort
from .. import API_FSA
import logging


class NLText(Resource):
    parser = reqparse.RequestParser()
    parser.add_argument("feature", type=str, default="Robot")
    parser.add_argument("feature_desc", type=str, default="")
    parser.add_argument("language", type=str, choices=("en", "cs"), required=True)
    parser.add_argument("background", type=str)
    parser.add_argument("text_raw", type=str, required=True)

    def get(self):
        if API_FSA.canRequestNLText():
            response = API_FSA.requestNLText()
            return response
        else:
            errorMessage = "NL text cannot be requested!"
            logging.error(errorMessage)
            abort(406, message=errorMessage)

    def post(self):
        args = self.parser.parse_args()

        try:
            response = API_FSA.receiveNLText(args)
        except Exception as error:
            errorMessage = "An error occurred while processing the NL text! Error: '{}'".format(str(error))
            logging.exception(errorMessage)
            abort(406, message=errorMessage)
        else:
            return response



# coding: utf-8
# from googletrans import Translator
import majka, re
from pattern.en import conjugate, parsetree, lemma
import spacy, unicodedata
import pkg_resources, yaml

from gherkan.containers.NLPModelManager import NLPModelManager
from gherkan.utils.czech_morpho import CZMorph
from gherkan.utils import logging_types
import logging


class Finetuner:
    def __init__(self, lang: str, nmm: NLPModelManager):
        self.nlp_model_en = nmm.nlp_model_en
        self.nlp_model_cs = CZMorph(nmm.nlp_model_cs)
        self.morph_lwt = nmm.morph_lwt
        self.morph_wlt = nmm.morph_wlt

        with pkg_resources.resource_stream("gherkan", "utils/verbs.yaml") as stream:
            self.verb_dic = yaml.full_load(stream)

    def conjugate(self, lang, word, person, number="singular"):
        """  Returns provided noun in given grammatical case / Vrátí podstatné jméno vyskloňované v daném pádu """
        negation = self.get_txt_info(word)[0]["negate"]
        if len(word.split()) > 1:
            return word
        if negation == "A":
            neg = False
        else:
            neg = True
        declenation = ""
        if lang == "en":
           try:
            if number == "singular":
                tag = str(person) + "sg"
                declenation = conjugate(word, tag)
            elif number == "plural":
                tag = str(person) + "pl"
                declenation = conjugate(word, tag)
           except:
               declenation = word
        elif lang == "cs":
           try:
            self.morph_lwt.flags |= majka.ADD_DIACRITICS
            analysis = self.morph_lwt.find(word)
            if analysis == []:
                declenation = word
            if number == "singular":
                for n in range(len(analysis)):
                    if 'singular' in analysis[n]['tags']:
                        declen = [analysis[n]['lemma'], analysis[n]['tags']]
                        if "person" in declen[1]:
                            if declen[1]["person"] == int(person) and declen[1]["singular"] is True and declen[1]["negation"] is neg:
                                declenation = declen[0]
                        elif "case" in declen[1]:
                            if declen[1]["case"] == int(person) and declen[1]["singular"] is True:
                                if "negation" in declen[1] and declen[1]["negation"] is not neg:
                                    pass
                                else:
                                    declenation = declen[0]
                                    break
            else:
                for n in range(len(analysis)):
                    if 'singular' in analysis[n]['tags']:
                        declen = [analysis[n]['lemma'], analysis[n]['tags']]
                        if "person" in declen[1]:
                            if declen[1]["person"] == int(person) and declen[1]["singular"] is False and declen[1]["negation"] is neg:
                                declenation = declen[0]
                        elif "case" in declen[1]:
                            if declen[1]["case"] == int(person) and declen[1]["singular"] is False:
                                declenation = declen[0]
           except:
               declenation = word
        if not declenation:
            declenation = word
        if word == "nebýt" and person == 3:   # majka does not work for this case
            if number == "singular":
                declenation = "není"
            else:
                declenation = "nejsou"
        return declenation

    def start_sentence_upper(self, text):
        """Changes the letters at the beginning of sentences into upper"""
        return re.sub("(^|[.?!])\s*([a-zA-Z])", lambda p: p.group(0).upper(), text.lower())

    def strip_extra_spaces(self, text):
        stripped_spaces = re.sub(' +', ' ', text)
        stripped_text = stripped_spaces.strip()
        return stripped_text

    def lemmatize(self, lang, word):
        """  Returns lemma of given word form / Vrátí základní podobu slova (např. infinitiv)"""
        lemma = word
        if lang == "cs":
            verb = self.find_verb(self.lang, word)
            if verb:
                for key in self.verb_dic[self.lang]["verbs"]:
                    if verb[0] in key.keys():
                        lemma = key[verb[0]]
                        return lemma
            self.morph_wlt.flags |= majka.ADD_DIACRITICS  # find word forms with diacritics
            self.morph_wlt.tags = False  # return just the lemma, do not process the tags
            self.morph_wlt.first_only = True  # return only the first entry
            output = self.morph_wlt.find(word)
            if output:
              lemma = output[0]['lemma']
        else:
            doc = self.nlp_model_en(word)
            for token in doc:
                    lemma = str(token.lemma_)

        return lemma

    def lemmatize_sentence(self, lang, text):
        with pkg_resources.resource_stream("gherkan", "utils/verbs.yaml") as stream:
            verb_dic = yaml.full_load(stream)
        with pkg_resources.resource_stream("gherkan", "utils/lemma_exceptions.yaml") as stream:
            lemma_dic = yaml.full_load(stream)
        self.lang = lang
        lemmas = ""
        skip_list = [",", ".", ";"]
        if self.lang == "cs":
            self.morph_wlt.flags |= majka.ADD_DIACRITICS  # find word forms with diacritics
            self.morph_wlt.tags = False  # return just the lemma, do not process the tags
            self.morph_wlt.first_only = True  # return only the first entry
            for word in text.split():
               if word in skip_list:
                   pass
               else:
                   found = False
                   for key in lemma_dic[self.lang]:
                       for keyword in lemma_dic[self.lang][key]:
                           if word.lower() == keyword:
                               lemmas += " "
                               lemmas += key
                               found = True
                               break
                   if not found:
                       output = self.morph_wlt.find(word)
                       lemmas += " "
                       if output:
                           lemma_det = output[0]['lemma']
                           lemmas += lemma_det
                       else:
                           lemmas += word



        elif self.lang == "en":
            # this version keeps capitals
            tree = parsetree(text, tokenize=True)

            for sentence in tree:
                for word in sentence:
                    if word.string in skip_list:
                        pass
                    found = False
                    for key in lemma_dic[self.lang]:
                        for keyword in lemma_dic[self.lang][key]:
                            if word.string.lower() == keyword:
                                lemmas += " "
                                lemmas += key
                                found = True
                                break
                    if not found:
                        lemmas += " "
                        lemmas += lemma(word.string)

        return lemmas

    def change_case(self, word, case, plural = False):
        """  Returns provided noun in given grammatical case / Vrátí podstatné jméno vyskloňované v daném pádu """
        self.morph_lwt.flags |= majka.ADD_DIACRITICS
        word_form = []
        analysis = self.morph_lwt.find(word)

        if plural == False:
            for n in range(len(analysis)):
                if 'singular' in analysis[n]['tags']:
                    declen = [analysis[n]['lemma'],analysis[n]['tags']['case']]
                    if case in declen:
                       word_form.append(declen[0])
                else:
                    pass
        else:
            for n in range(len(analysis)):
                if 'plural' in analysis[n]['tags']:
                    declen = [analysis[n]['lemma'], analysis[n]['tags']['case']]
                    if case in declen:
                        word_form.append(declen[0])
            else:
                pass

        return word_form[0]

    def make_negative(self, lang, word):
        if lang == "cs":
            self.morph_wlt.flags |= majka.ADD_DIACRITICS
            self.morph_wlt.tags = False
            self.morph_wlt.first_only = False
            analysis = self.morph_wlt.find(word)
        elif lang == "en":
            neg = conjugate(word, negated=True)

    def get_cases(self, word):
        """ Returns all possible cases of the word, in case of verb returns all info / Vrátí možné pády daného slova, u sloves vrátí všechno info"""
        self.morph_wlt.flags |= majka.ADD_DIACRITICS  # find word forms with diacritics
        output = self.morph_wlt.find(word)

        cases = []
        if "substantive" in output[0]['tags']['pos'] or "adjective" in output[0]['tags']['pos']:
            for n in range(len(output)):
                 case = output[n]['tags']['case']
                 if "singular" in output[n]['tags']:
                  number = "singular"
                 else:
                  number = "plural"
                 cases.append([case, number])

        elif "verb" in output[0]['tags']['pos']:
            cases = output

        return cases


    def align_adjectives(self, lang, sentence):
        word_list = sentence.split()
        self.morph_wlt.flags |= majka.ADD_DIACRITICS
        config = []
        if lang =="cs":
            for word in word_list:
                analysis = self.morph_wlt.find(word)
                if analysis == []:
                   continue
                else:
                    if 'tags' in analysis[0]:
                        if analysis[0]["tags"]["pos"] == "substantive":
                            config = analysis[0]["tags"]
                            config["negation"] = False
                            del config["pos"]
                            if 'plural' in config:
                                del config["plural"]
            if config:
                verb = self.find_verb(lang, sentence)
                try:
                   analysis = self.morph_wlt.find(verb[0])
                   config["singular"] = analysis[0]["tags"]["singular"]
                except:
                    pass
                for word in word_list:
                    analysis = self.morph_lwt.find(word)
                    if analysis:
                        if analysis[0]["tags"]["pos"] == "adjective":
                            for dict in analysis:
                                if all(item in dict["tags"].items() for item in config.items()):
                                    aligned_adj = dict["lemma"]
                                    sentence = re.sub(word, aligned_adj, sentence)
        else:
            pass
        return sentence

    def remove_diacritics(self, text):
        """
        Returns a string with all diacritics (aka non-spacing marks) removed.
        For example "Héllô" will become "Hello".
        Useful for comparing strings in an accent-insensitive fashion.
        """
        normalized = unicodedata.normalize("NFKD", text)
        return "".join(c for c in normalized if unicodedata.category(c) != "Mn")

    def get_txt_info(self, text):
        """For Czech only - get POS tagging from Morphodita"""
        text_info = self.nlp_model_cs.get_pos_info(text)
        return text_info

    def find_verb(self, lang, sentence, head_only=True):
        verb = []
        word_list = sentence.split()
        if lang =="cs":
            for word in word_list:
                analysis = self.morph_wlt.find(word)
                if analysis == []:
                   continue
                else:
                    if 'tags' in analysis[0]:
                        if analysis[0]["tags"]["pos"] == "verb":
                            verb.append(word)
                        else:
                            continue
            if not verb:
                sent_lst = self.nlp_model_cs.get_pos_info(sentence)
                for x in sent_lst:
                    if x["pos"] == "V":
                        verb.append(x["word"])

        elif lang == "en":
            t = parsetree(sentence, tokenize=True,  tags=True)
            excepts = ["start", "stop", "end", "finish"]
            for sent in t:
                  chunk = sent.verbs
                  check = False
                  if chunk:
                      for x in excepts:
                              if x in chunk[0].string:
                                word = re.findall(r'\w*{}\w*'.format(x), chunk[0].string)
                                if word:
                                  verb.append(word[0])
                                  check = True
                      if not check:
                             if head_only:
                               verb.append(chunk[0].head.string)
                             else:
                               verb.append(chunk[0].string)
                  else:
                    for x in t.string.split():
                      if x in excepts:
                          word = re.findall(r'\w*{}\w*'.format(x), t.string)
                          if word:
                              verb.append(word[0])
        sentence_l = self.lemmatize_sentence(lang, sentence)
        if verb:
            verb_copy = verb
            for v in verb_copy:
              if v in self.verb_dic[lang]["not verbs"]:
                verb.remove(v)
              if not verb:
                new_phrase = sentence.replace(v, "")
                self.find_verb(lang, new_phrase)
        if not verb:
            for word in sentence_l.split():
                if word in self.verb_dic[lang]["verbs"]:
                    verb = [word.replace("-", "ne")]
        if not verb:
            logging.debug("Warning: did not detect any verb in sentence: {}", sentence)
        return verb

    # def translate(self, phrase, lang = "cz-en"):
    #     """ Translates a word or phrase from czech to english (lang = "cz-en") or from english to czech (lang = "en-cz") """
    #     translator = Translator()
    #     if lang == "cz-en":
    #       result = translator.translate(phrase, src="cs", dest='en')
    #     elif lang == "en-cz":
    #       result = translator.translate(phrase, src="en", dest='cs')
    #
    #     translation = result.text
    #     return translation






from gherkan.utils import constants as c

FEATURE = "Feature"
BACKGROUND = "Background"
SCENARIO = "Scenario"
GIVEN = "Given"
WHEN = "When"
THEN = "Then"
AND = "And"


FEATURE_CS = "Požadavek"
BACKGROUND_CS = "Kontext"
SCENARIO_CS = "Scénář"
GIVEN_CS = "Pokud"
WHEN_CS = "Když"
THEN_CS = "Pak"
AND_CS = "A"

KEYWORDS_EN = {
    "feature": FEATURE,
    "background": BACKGROUND,
    "scenario": SCENARIO,
    "given": GIVEN,
    "when": WHEN,
    "then": THEN,
    "and": AND,
}

KEYWORDS_CS = {
    "feature": FEATURE_CS,
    "background": BACKGROUND_CS,
    "scenario": SCENARIO_CS,
    "given": GIVEN_CS,
    "when": WHEN_CS,
    "then": THEN_CS,
    "and": AND_CS,
}

def get_kw(lang : str, key : str):
    if lang == c.LANG_EN:
        return KEYWORDS_EN[key.lower()]
    elif lang == c.LANG_CZ:
        return  KEYWORDS_CS[key.lower()]
    else:
        raise ValueError(f"Language not recognized: {lang}")


from flask import jsonify, request, Response
from flask_restful import Resource, reqparse, abort
import os
from .. import API_FSA
import logging


class Audio(Resource):
    parser = reqparse.RequestParser()
    parser.add_argument("language", type=str, choices=("en", "cs"), required=True)

    def post(self):
        """
        request = {
            "language": "en",
            "description": "desc text",
            "background", "optional background text"
            }
        E.g.: curl -F file=@nl_audio_file.wav -F 'data={"language": "en", "description": "desc text", "background", ""}'  <URI>/audio

        response = {
            "language": "en",
            "description": "desc text",
            "background", "optional background text",
            "transcript": "text extracted from audio"
        }
        """
        args = self.parser.parse_args()

        for _, value in request.files.items():
            audioFileName = os.path.join(API_FSA.audioFolder, value.filename)
            value.save(audioFileName)
        try:
            transcript = API_FSA.receiveAudio(audioFileName, args["language"])
        except Exception as error:
            errorMessage = "An error occurred while processing the audio! Error: '{}'".format(str(error))
            logging.exception(errorMessage)
            abort(400, message=errorMessage)
        else:
            return {
                "language": args["language"],
                "transcript": transcript
                }





class LexerConf:
    def __init__(self, tokens, ignore=(), postlex=None, callbacks=None):
        self.tokens = tokens
        self.ignore = ignore
        self.postlex = postlex
        self.callbacks = callbacks or {}

class ParserConf:
    def __init__(self, rules, callback, start):
        self.rules = rules
        self.callback = callback
        self.start = start






class SignalPhrase:
    def __init__(self):
        self.language = None
        self.tempName = None
        self.tempStr = None
        self.tempToNL = None
        self.inclinations = None
        self.negate = False
        self.force = False
        self.edge = False
        self.vars = {}
        self.values = {}
        self.niceStr = None


# The file was automatically generated by Lark v0.5.5
#
#
#   Lark Stand-alone Generator Tool
# ----------------------------------
# Generates a stand-alone LALR(1) parser with a standard lexer
#
# Git:    https://github.com/erezsh/lark
# Author: Erez Shinan (erezshin@gmail.com)
#
#
#    >>> LICENSE
#
#    This tool and its generated code use a separate license from Lark.
#
#    It is licensed under GPLv2 or above.
#
#    If you wish to purchase a commercial license for this tool and its
#    generated code, contact me via email.
#
#    This program is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 2 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    See <http://www.gnu.org/licenses/>.
#
#


import types
import functools
from contextlib import contextmanager

Str = type(u'')

def inline_args(f):
    # print '@@', f.__name__, type(f), isinstance(f, types.FunctionType), isinstance(f, types.TypeType), isinstance(f, types.BuiltinFunctionType)
    if isinstance(f, types.FunctionType):
        @functools.wraps(f)
        def _f_func(self, args):
            return f(self, *args)
        return _f_func
    elif isinstance(f, (type, types.BuiltinFunctionType)):
        @functools.wraps(f)
        def _f_builtin(_self, args):
            return f(*args)
        return _f_builtin
    elif isinstance(f, types.MethodType):
        @functools.wraps(f.__func__)
        def _f(self, args):
            return f.__func__(self, *args)
        return _f
    else:
        @functools.wraps(f.__call__.__func__)
        def _f(self, args):
            return f.__call__.__func__(self, *args)
        return _f


try:
    from contextlib import suppress     # Python 3
except ImportError:
    @contextmanager
    def suppress(*excs):
        '''Catch and dismiss the provided exception

        >>> x = 'hello'
        >>> with suppress(IndexError):
        ...     x = x[10]
        >>> x
        'hello'
        '''
        try:
            yield
        except excs:
            pass


def is_terminal(sym):
    return sym.isupper()

class GrammarError(Exception):
    pass

class ParseError(Exception):
    pass

class UnexpectedToken(ParseError):
    def __init__(self, token, expected, seq, index, considered_rules=None):
        self.token = token
        self.expected = expected
        self.line = getattr(token, 'line', '?')
        self.column = getattr(token, 'column', '?')
        self.considered_rules = considered_rules

        try:
            context = ' '.join(['%r(%s)' % (t.value, t.type) for t in seq[index:index+5]])
        except AttributeError:
            context = seq[index:index+5]
        except TypeError:
            context = "<no context>"
        message = ("Unexpected token %r at line %s, column %s.\n"
                   "Expected: %s\n"
                   "Context: %s" % (token, self.line, self.column, expected, context))

        super(UnexpectedToken, self).__init__(message)



class Tree(object):
    def __init__(self, data, children):
        self.data = data
        self.children = children

    def __repr__(self):
        return 'Tree(%s, %s)' % (self.data, self.children)

    def _pretty_label(self):
        return self.data

    def _pretty(self, level, indent_str):
        if len(self.children) == 1 and not isinstance(self.children[0], Tree):
            return [ indent_str*level, self._pretty_label(), '\t', '%s' % (self.children[0],), '\n']

        l = [ indent_str*level, self._pretty_label(), '\n' ]
        for n in self.children:
            if isinstance(n, Tree):
                l += n._pretty(level+1, indent_str)
            else:
                l += [ indent_str*(level+1), '%s' % (n,), '\n' ]

        return l

    def pretty(self, indent_str='  '):
        return ''.join(self._pretty(0, indent_str))
class Transformer(object):
    def _get_func(self, name):
        return getattr(self, name)

    def transform(self, tree):
        items = []
        for c in tree.children:
            try:
                items.append(self.transform(c) if isinstance(c, Tree) else c)
            except Discard:
                pass
        try:
            f = self._get_func(tree.data)
        except AttributeError:
            return self.__default__(tree.data, items)
        else:
            return f(items)

    def __default__(self, data, children):
        return Tree(data, children)

    def __mul__(self, other):
        return TransformerChain(self, other)


class Discard(Exception):
    pass

class TransformerChain(object):
    def __init__(self, *transformers):
        self.transformers = transformers

    def transform(self, tree):
        for t in self.transformers:
            tree = t.transform(tree)
        return tree

    def __mul__(self, other):
        return TransformerChain(*self.transformers + (other,))



class InlineTransformer(Transformer):
    def _get_func(self, name):  # use super()._get_func
        return inline_args(getattr(self, name)).__get__(self)


class Visitor(object):
    def visit(self, tree):
        for child in tree.children:
            if isinstance(child, Tree):
                self.visit(child)

        f = getattr(self, tree.data, self.__default__)
        f(tree)
        return tree

    def __default__(self, tree):
        pass


class Visitor_NoRecurse(Visitor):
    def visit(self, tree):
        subtrees = list(tree.iter_subtrees())

        for subtree in (subtrees):
            getattr(self, subtree.data, self.__default__)(subtree)
        return tree


class Transformer_NoRecurse(Transformer):
    def transform(self, tree):
        subtrees = list(tree.iter_subtrees())

        def _t(t):
            # Assumes t is already transformed
            try:
                f = self._get_func(t.data)
            except AttributeError:
                return self.__default__(t)
            else:
                return f(t)

        for subtree in subtrees:
            children = []
            for c in subtree.children:
                try:
                    children.append(_t(c) if isinstance(c, Tree) else c)
                except Discard:
                    pass
            subtree.children = children

        return _t(tree)

    def __default__(self, t):
        return t

class Indenter:
    def __init__(self):
        self.paren_level = 0
        self.indent_level = [0]

    def handle_NL(self, token):
        if self.paren_level > 0:
            return

        yield token

        indent_str = token.rsplit('\n', 1)[1] # Tabs and spaces
        indent = indent_str.count(' ') + indent_str.count('\t') * self.tab_len

        if indent > self.indent_level[-1]:
            self.indent_level.append(indent)
            yield Token.new_borrow_pos(self.INDENT_type, indent_str, token)
        else:
            while indent < self.indent_level[-1]:
                self.indent_level.pop()
                yield Token.new_borrow_pos(self.DEDENT_type, indent_str, token)

            assert indent == self.indent_level[-1], '%s != %s' % (indent, self.indent_level[-1])

    def process(self, stream):
        for token in stream:
            if token.type == self.NL_type:
                for t in self.handle_NL(token):
                    yield t
            else:
                yield token

            if token.type in self.OPEN_PAREN_types:
                self.paren_level += 1
            elif token.type in self.CLOSE_PAREN_types:
                self.paren_level -= 1
                assert self.paren_level >= 0

        while len(self.indent_level) > 1:
            self.indent_level.pop()
            yield Token(self.DEDENT_type, '')

        assert self.indent_level == [0], self.indent_level

    # XXX Hack for ContextualLexer. Maybe there's a more elegant solution?
    @property
    def always_accept(self):
        return (self.NL_type,)


class LexError(Exception):
    pass

class UnexpectedInput(LexError):
    def __init__(self, seq, lex_pos, line, column, allowed=None, considered_rules=None):
        context = seq[lex_pos:lex_pos+5]
        message = "No token defined for: '%s' in %r at line %d col %d" % (seq[lex_pos], context, line, column)
        if allowed:
            message += '\n\nExpecting: %s\n' % allowed

        super(UnexpectedInput, self).__init__(message)

        self.line = line
        self.column = column
        self.context = context
        self.allowed = allowed
        self.considered_rules = considered_rules

class Token(Str):
    def __new__(cls, type_, value, pos_in_stream=None, line=None, column=None):
        inst = Str.__new__(cls, value)
        inst.type = type_
        inst.pos_in_stream = pos_in_stream
        inst.value = value
        inst.line = line
        inst.column = column
        return inst

    @classmethod
    def new_borrow_pos(cls, type_, value, borrow_t):
        return cls(type_, value, borrow_t.pos_in_stream, line=borrow_t.line, column=borrow_t.column)

    def __repr__(self):
        return 'Token(%s, %r)' % (self.type, self.value)

    def __deepcopy__(self, memo):
        return Token(self.type, self.value, self.pos_in_stream, self.line, self.column)

    def __eq__(self, other):
        if isinstance(other, Token) and self.type != other.type:
            return False

        return Str.__eq__(self, other)

    __hash__ = Str.__hash__


class LineCounter:
    def __init__(self):
        self.newline_char = '\n'
        self.char_pos = 0
        self.line = 1
        self.column = 0
        self.line_start_pos = 0

    def feed(self, token, test_newline=True):
        """Consume a token and calculate the new line & column.

        As an optional optimization, set test_newline=False is token doesn't contain a newline.
        """
        if test_newline:
            newlines = token.count(self.newline_char)
            if newlines:
                self.line += newlines
                self.line_start_pos = self.char_pos + token.rindex(self.newline_char) + 1

        self.char_pos += len(token)
        self.column = self.char_pos - self.line_start_pos

class _Lex:
    "Built to serve both Lexer and ContextualLexer"
    def __init__(self, lexer):
        self.lexer = lexer

    def lex(self, stream, newline_types, ignore_types):
        newline_types = list(newline_types)
        ignore_types = list(ignore_types)
        line_ctr = LineCounter()

        t = None
        while True:
            lexer = self.lexer
            for mre, type_from_index in lexer.mres:
                m = mre.match(stream, line_ctr.char_pos)
                if m:
                    value = m.group(0)
                    type_ = type_from_index[m.lastindex]
                    if type_ not in ignore_types:
                        t = Token(type_, value, line_ctr.char_pos, line_ctr.line, line_ctr.column)
                        if t.type in lexer.callback:
                            t = lexer.callback[t.type](t)
                        yield t
                    else:
                        if type_ in lexer.callback:
                            t = Token(type_, value, line_ctr.char_pos, line_ctr.line, line_ctr.column)
                            lexer.callback[type_](t)

                    line_ctr.feed(value, type_ in newline_types)
                    if t:
                        t.end_line = line_ctr.line
                        t.end_column = line_ctr.column
                    break
            else:
                if line_ctr.char_pos < len(stream):
                    raise UnexpectedInput(stream, line_ctr.char_pos, line_ctr.line, line_ctr.column)
                break

class UnlessCallback:
    def __init__(self, mres):
        self.mres = mres

    def __call__(self, t):
        for mre, type_from_index in self.mres:
            m = mre.match(t.value)
            if m:
                value = m.group(0)
                t.type = type_from_index[m.lastindex]
                break
        return t



from functools import partial

class ExpandSingleChild:
    def __init__(self, node_builder):
        self.node_builder = node_builder

    def __call__(self, children):
        if len(children) == 1:
            return children[0]
        else:
            return self.node_builder(children)


class CreateToken:
    "Used for fixing the results of scanless parsing"

    def __init__(self, token_name, node_builder):
        self.node_builder = node_builder
        self.token_name = token_name

    def __call__(self, children):
        return self.node_builder( [Token(self.token_name, ''.join(children))] )


class PropagatePositions:
    def __init__(self, node_builder):
        self.node_builder = node_builder

    def __call__(self, children):
        res = self.node_builder(children)

        if children:
            for a in children:
                with suppress(AttributeError):
                    res.line = a.line
                    res.column = a.column
                break

            for a in reversed(children):
                with suppress(AttributeError):
                    res.end_line = a.end_line
                    res.end_column = a.end_column
                break

        return res


class ChildFilter:
    def __init__(self, to_include, node_builder):
        self.node_builder = node_builder
        self.to_include = to_include

    def __call__(self, children):
        filtered = []
        for i, to_expand in self.to_include:
            if to_expand:
                if filtered:
                    filtered += children[i].children
                else:   # Optimize for left-recursion
                    filtered = children[i].children
            else:
                filtered.append(children[i])

        return self.node_builder(filtered)

def _should_expand(sym):
    return not is_terminal(sym) and sym.startswith('_')

def maybe_create_child_filter(expansion, filter_out):
    to_include = [(i, _should_expand(sym)) for i, sym in enumerate(expansion) if sym not in filter_out]

    if len(to_include) < len(expansion) or any(to_expand for i, to_expand in to_include):
        return partial(ChildFilter, to_include)


class Callback(object):
    pass

class ParseTreeBuilder:
    def __init__(self, rules, tree_class, propagate_positions=False, keep_all_tokens=False):
        self.tree_class = tree_class
        self.propagate_positions = propagate_positions
        self.always_keep_all_tokens = keep_all_tokens

        self.rule_builders = list(self._init_builders(rules))

        self.user_aliases = {}

    def _init_builders(self, rules):
        filter_out = {rule.origin for rule in rules if rule.options and rule.options.filter_out}
        filter_out |= {sym for rule in rules for sym in rule.expansion if is_terminal(sym) and sym.startswith('_')}
        assert all(x.startswith('_') for x in filter_out)

        for rule in rules:
            options = rule.options
            keep_all_tokens = self.always_keep_all_tokens or (options.keep_all_tokens if options else False)
            expand_single_child = options.expand1 if options else False
            create_token = options.create_token if options else False

            wrapper_chain = filter(None, [
                create_token and partial(CreateToken, create_token),
                (expand_single_child and not rule.alias) and ExpandSingleChild,
                maybe_create_child_filter(rule.expansion, () if keep_all_tokens else filter_out),
                self.propagate_positions and PropagatePositions,
            ])

            yield rule, wrapper_chain


    def create_callback(self, transformer=None):
        callback = Callback()

        for rule, wrapper_chain in self.rule_builders:
            internal_callback_name = '_callback_%s_%s' % (rule.origin, '_'.join(rule.expansion))

            user_callback_name = rule.alias or rule.origin
            try:
                f = transformer._get_func(user_callback_name)
            except AttributeError:
                f = partial(self.tree_class, user_callback_name)

            self.user_aliases[rule] = rule.alias
            rule.alias = internal_callback_name

            for w in wrapper_chain:
                f = w(f)

            if hasattr(callback, internal_callback_name):
                raise GrammarError("Rule '%s' already exists" % (rule,))
            setattr(callback, internal_callback_name, f)

        return callback



class _Parser:
    def __init__(self, parse_table, callbacks):
        self.states = parse_table.states
        self.start_state = parse_table.start_state
        self.end_state = parse_table.end_state
        self.callbacks = callbacks

    def parse(self, seq, set_state=None):
        i = 0
        token = None
        stream = iter(seq)
        states = self.states

        state_stack = [self.start_state]
        value_stack = []

        if set_state: set_state(self.start_state)

        def get_action(key):
            state = state_stack[-1]
            try:
                return states[state][key]
            except KeyError:
                expected = states[state].keys()

                raise UnexpectedToken(token, expected, seq, i)

        def reduce(rule):
            size = len(rule.expansion)
            if size:
                s = value_stack[-size:]
                del state_stack[-size:]
                del value_stack[-size:]
            else:
                s = []

            value = self.callbacks[rule](s)

            _action, new_state = get_action(rule.origin)
            assert _action is Shift
            state_stack.append(new_state)
            value_stack.append(value)

        # Main LALR-parser loop
        for i, token in enumerate(stream):
            while True:
                action, arg = get_action(token.type)
                assert arg != self.end_state

                if action is Shift:
                    state_stack.append(arg)
                    value_stack.append(token)
                    if set_state: set_state(arg)
                    break # next token
                else:
                    reduce(arg)

        while True:
            _action, arg = get_action('$END')
            if _action is Shift:
                assert arg == self.end_state
                val ,= value_stack
                return val
            else:
                reduce(arg)



class Rule(object):
    """
        origin : a symbol
        expansion : a list of symbols
    """
    def __init__(self, origin, expansion, alias=None, options=None):
        self.origin = origin
        self.expansion = expansion
        self.alias = alias
        self.options = options

    def __str__(self):
        return '<%s : %s>' % (self.origin, ' '.join(map(str,self.expansion)))

    def __repr__(self):
        return 'Rule(%r, %r, %r, %r)' % (self.origin, self.expansion, self.alias, self.options)


class RuleOptions:
    def __init__(self, keep_all_tokens=False, expand1=False, create_token=None, filter_out=False, priority=None):
        self.keep_all_tokens = keep_all_tokens
        self.expand1 = expand1
        self.create_token = create_token  # used for scanless postprocessing
        self.priority = priority

        self.filter_out = filter_out        # remove this rule from the tree
                                            # used for "token"-rules in scanless

    def __repr__(self):
        return 'RuleOptions(%r, %r, %r, %r, %r)' % (
            self.keep_all_tokens,
            self.expand1,
            self.create_token,
            self.priority,
            self.filter_out
        )

Shift = 0
Reduce = 1
import re
MRES = (
[(u'(?P<SIGNED_NUMBER>(?:(?:\\+|\\-))?(?:(?:(?:[0-9])+(?:e|E)(?:(?:\\+|\\-))?(?:[0-9])+|(?:(?:[0-9])+\\.(?:(?:[0-9])+)?|\\.(?:[0-9])+)(?:(?:e|E)(?:(?:\\+|\\-))?(?:[0-9])+)?)|(?:[0-9])+))|(?P<ESCAPED_STRING>\\"(?:(?:\\\\\\"|[^"]))*\\")|(?P<WS>(?:[ \t\x0c\r\n])+)|(?P<__FALSE1>false)|(?P<__NULL2>null)|(?P<__TRUE0>true)|(?P<__COLON>\\:)|(?P<__COMMA>\\,)|(?P<__LBRACE>\\{)|(?P<__LSQB>\\[)|(?P<__RBRACE>\\})|(?P<__RSQB>\\])',
  {1: u'SIGNED_NUMBER',
   2: u'ESCAPED_STRING',
   3: u'WS',
   4: u'__FALSE1',
   5: u'__NULL2',
   6: u'__TRUE0',
   7: u'__COLON',
   8: u'__COMMA',
   9: u'__LBRACE',
   10: u'__LSQB',
   11: u'__RBRACE',
   12: u'__RSQB'})]
)
LEXER_CALLBACK = (
{}
)
NEWLINE_TYPES = [u'WS']
IGNORE_TYPES = [u'WS']
class LexerRegexps: pass
lexer_regexps = LexerRegexps()
lexer_regexps.mres = [(re.compile(p), d) for p, d in MRES]
lexer_regexps.callback = {n: UnlessCallback([(re.compile(p), d) for p, d in mres])
                          for n, mres in LEXER_CALLBACK.items()}
lexer = _Lex(lexer_regexps)
def lex(stream):
    return lexer.lex(stream, NEWLINE_TYPES, IGNORE_TYPES)
RULES = {
  0: Rule(u'start', [u'value'], None, RuleOptions(False, True, None, None, False)),
  1: Rule(u'value', [u'string'], None, RuleOptions(False, True, None, None, False)),
  2: Rule(u'value', [u'__TRUE0'], u'true', RuleOptions(False, True, None, None, False)),
  3: Rule(u'value', [u'array'], None, RuleOptions(False, True, None, None, False)),
  4: Rule(u'value', [u'__NULL2'], u'null', RuleOptions(False, True, None, None, False)),
  5: Rule(u'value', [u'SIGNED_NUMBER'], u'number', RuleOptions(False, True, None, None, False)),
  6: Rule(u'value', [u'object'], None, RuleOptions(False, True, None, None, False)),
  7: Rule(u'value', [u'__FALSE1'], u'false', RuleOptions(False, True, None, None, False)),
  8: Rule(u'array', ['__LSQB', u'value', '__RSQB'], None, RuleOptions(False, False, None, None, False)),
  9: Rule(u'array', ['__LSQB', u'value', '__anon_star_0', '__RSQB'], None, RuleOptions(False, False, None, None, False)),
  10: Rule(u'array', ['__LSQB', '__RSQB'], None, RuleOptions(False, False, None, None, False)),
  11: Rule(u'object', ['__LBRACE', u'pair', '__anon_star_1', '__RBRACE'], None, RuleOptions(False, False, None, None, False)),
  12: Rule(u'object', ['__LBRACE', '__RBRACE'], None, RuleOptions(False, False, None, None, False)),
  13: Rule(u'object', ['__LBRACE', u'pair', '__RBRACE'], None, RuleOptions(False, False, None, None, False)),
  14: Rule(u'pair', [u'string', '__COLON', u'value'], None, RuleOptions(False, False, None, None, False)),
  15: Rule(u'string', [u'ESCAPED_STRING'], None, RuleOptions(False, False, None, None, False)),
  16: Rule('__anon_star_0', ['__anon_star_0', '__COMMA', u'value'], None, None),
  17: Rule('__anon_star_0', ['__COMMA', u'value'], None, None),
  18: Rule('__anon_star_1', ['__COMMA', u'pair'], None, None),
  19: Rule('__anon_star_1', ['__anon_star_1', '__COMMA', u'pair'], None, None),
}
parse_tree_builder = ParseTreeBuilder(RULES.values(), Tree)
class ParseTable: pass
parse_table = ParseTable()
STATES = {
  0: {0: (1, 4), 1: (1, 4), 2: (1, 4), 3: (1, 4)},
  1: {1: (1, 14), 2: (1, 14)},
  2: {0: (0, 29), 1: (0, 32), 4: (0, 9)},
  3: {1: (0, 13), 2: (0, 12)},
  4: {0: (1, 1), 1: (1, 1), 2: (1, 1), 3: (1, 1)},
  5: {0: (1, 10), 1: (1, 10), 2: (1, 10), 3: (1, 10)},
  6: {2: (0, 15), 5: (0, 27), 6: (0, 16), 7: (0, 26)},
  7: {5: (0, 34), 6: (0, 16), 7: (0, 26)},
  8: {0: (1, 2), 1: (1, 2), 2: (1, 2), 3: (1, 2)},
  9: {0: (0, 11), 1: (0, 22)},
  10: {0: (1, 6), 1: (1, 6), 2: (1, 6), 3: (1, 6)},
  11: {0: (1, 9), 1: (1, 9), 2: (1, 9), 3: (1, 9)},
  12: {0: (1, 11), 1: (1, 11), 2: (1, 11), 3: (1, 11)},
  13: {5: (0, 20), 6: (0, 16), 7: (0, 26)},
  14: {6: (0, 16), 7: (0, 4), 8: (0, 6), 9: (0, 31), 10: (0, 24), 11: (0, 10), 12: (0, 21), 13: (0, 17), 14: (0, 33), 15: (0, 0), 16: (0, 19), 17: (0, 8)},
  15: {0: (1, 12), 1: (1, 12), 2: (1, 12), 3: (1, 12)},
  16: {0: (1, 15), 1: (1, 15), 2: (1, 15), 3: (1, 15), 18: (1, 15)},
  17: {3: (1, 0)},
  18: {},
  19: {0: (1, 3), 1: (1, 3), 2: (1, 3), 3: (1, 3)},
  20: {1: (1, 19), 2: (1, 19)},
  21: {0: (1, 5), 1: (1, 5), 2: (1, 5), 3: (1, 5)},
  22: {6: (0, 16), 7: (0, 4), 8: (0, 6), 9: (0, 31), 10: (0, 24), 11: (0, 10), 12: (0, 21), 13: (0, 30), 15: (0, 0), 16: (0, 19), 17: (0, 8)},
  23: {6: (0, 16), 7: (0, 4), 8: (0, 6), 9: (0, 31), 10: (0, 24), 11: (0, 10), 12: (0, 21), 13: (0, 1), 15: (0, 0), 16: (0, 19), 17: (0, 8)},
  24: {0: (0, 5), 6: (0, 16), 7: (0, 4), 8: (0, 6), 9: (0, 31), 10: (0, 24), 11: (0, 10), 12: (0, 21), 13: (0, 2), 15: (0, 0), 16: (0, 19), 17: (0, 8)},
  25: {0: (1, 13), 1: (1, 13), 2: (1, 13), 3: (1, 13)},
  26: {18: (0, 23)},
  27: {1: (0, 7), 2: (0, 25), 19: (0, 3)},
  28: {0: (1, 17), 1: (1, 17)},
  29: {0: (1, 8), 1: (1, 8), 2: (1, 8), 3: (1, 8)},
  30: {0: (1, 16), 1: (1, 16)},
  31: {0: (1, 7), 1: (1, 7), 2: (1, 7), 3: (1, 7)},
  32: {6: (0, 16), 7: (0, 4), 8: (0, 6), 9: (0, 31), 10: (0, 24), 11: (0, 10), 12: (0, 21), 13: (0, 28), 15: (0, 0), 16: (0, 19), 17: (0, 8)},
  33: {3: (0, 18)},
  34: {1: (1, 18), 2: (1, 18)},
}
TOKEN_TYPES = (
{0: '__RSQB',
 1: '__COMMA',
 2: '__RBRACE',
 3: '$END',
 4: '__anon_star_0',
 5: u'pair',
 6: u'ESCAPED_STRING',
 7: u'string',
 8: '__LBRACE',
 9: u'__FALSE1',
 10: '__LSQB',
 11: u'object',
 12: u'SIGNED_NUMBER',
 13: u'value',
 14: 'start',
 15: u'__NULL2',
 16: u'array',
 17: u'__TRUE0',
 18: '__COLON',
 19: '__anon_star_1'}
)
parse_table.states = {s: {TOKEN_TYPES[t]: (a, RULES[x] if a is Reduce else x) for t, (a, x) in acts.items()}
                      for s, acts in STATES.items()}
parse_table.start_state = 14
parse_table.end_state = 18
class Lark_StandAlone:
  def __init__(self, transformer=None, postlex=None):
     callback = parse_tree_builder.create_callback(transformer=transformer)
     callbacks = {rule: getattr(callback, rule.alias or rule.origin, None) for rule in RULES.values()}
     self.parser = _Parser(parse_table, callbacks)
     self.postlex = postlex
  def parse(self, stream):
     tokens = lex(stream)
     if self.postlex: tokens = self.postlex.process(tokens)
     return self.parser.parse(tokens)



import os
import gherkan.utils.constants as c

from gherkan.flask_api.raw_text_to_signal import nl_to_signal

# request = {
#     "feature": "robot R4",
#     "feature_desc": "Robot R4 je kolaborativní robot Kuka IIWA. Má za úkol finální kompletaci autíčka. Postupně mu jsou na jeho stanici XX přivezeny zkompletované součásti autíčka. Robotu je nejdříve přivezono chassi autíčka. Následuje body (korba) a jako poslední je přivezena kabina.",
#     "background": "given line is on",
#     "text_raw": "scenario   As soon as robot R1 finished sorting all cubes, shuttle goes to station XXX",
#     "language": "en"
# }

request = {
    "feature" : "Robot N2",
    "feature_desc" : "Po dokončení svařováhí je výrobek pomocí robota N2 přemístěn ze stolu na dopravník. Na doravníkový pás se může položit výrobek poze při splěnní násoledujících podmínke. Místo pro umístění výrobku na pás je volné, pás není v pohybu a lakování předchozího výrobku bylo úspoěšně ukončeno tzn. předchozí výrobek odjel z pozice pro lakování. Pás se nesmí rozjet v případě, že robot nedolakoval výrobek. Hrozí poškození robota a jeho lakovacích pistolí a výrobek by nebyl správně nalakován.",
    "background" : "pokud linka je v provozu",
    "text_raw" : "scénář  Jakmile robot N1 skončí svařit díl 1 s dílem 2 a robot N2 skončí položit produkt na pás a robot N2 je inaktivní, pak robot N2 začne zvedat produkt \n scénář  Pokud produkt není na začátku a produkt není na pozici a dopravník zastaví a robot N2 skončí zvedat produkt, pak robot N2 začne položit produkt na pás \n scénář  Pokud je produkt na začátku nebo je produkt na pozici nebo dopravník začne, jakmile robot N2 skončí zvedat produkt, pak robot nepoloží produkt na pás",
    "language" : "cs"
}


base_path = os.path.join(c.DATA_DIR, "input", "raw_out")

nl_to_signal(base_path, request)

nl_file_path = base_path + ".feature"
signal_file_path = base_path + "_signals.feature"

print("\n\n--- NL FILE ---")
with open(nl_file_path, "rt", encoding="utf-8") as f:
    text = f.read()
    print(text)

print("\n\n--- SIGNAL FILE ---")
with open(signal_file_path, "rt", encoding="utf-8") as f:
    text = f.read()
    print(text)


import requests
import unittest

import gherkan.utils.gherkin_keywords as g
import gherkan.utils.constants as c



class TestNLStatements(unittest.TestCase):

    def setUp(self):
        self.url = f"http://{self.__class__.host}:{self.__class__.port}/{{}}"

        resetResponse = requests.get(self.url.format("reset"))
        self.assertDictContainsSubset({"OK": True}, resetResponse.json())


    def compare_background_content(self, content, correct_background, lang):
        content_list = content.split("\n")
        content_list = [s.strip() for s in content_list]
        background_line_text = (g.get_kw(lang, "Background")) + ":"
        res_idx = content_list.index(background_line_text) + 1

        # case insensitive
        self.assertEqual(content_list[res_idx].lower(), correct_background.lower())


    def createStatementRequest(self, lang, statement):
        request = {
            "feature": "StatementTest",
            "feature_desc": "StatementTest dummy description",
            "language": lang,
            "background": statement,
            # dummy scenario, only background is tested
            "text_raw": "scenario given line is on" if lang == "en"
                            else "scenario pokud linka je zapnutá"
        }
        return request

    def singleStatementTest(self, lang, statement, correctResponse):
        request = self.createStatementRequest(lang, statement)

        response = requests.post(self.url.format("nltext"), data=request)
        try:
            response = response.json()
        except Exception as exception:
            self.fail(
                "An exception was raised while trying to convert response to JSON. "
                "This is most likely due to the fact that the response was of different type.\n"
                "The exception: {}\nResponse text (if possible):\n{}".format(
                    exception, response.text))
        else:
            self.assertDictContainsSubset({"errors": False}, response)

            response = requests.get(self.url.format("signals"))

            self.compare_background_content(response.content.decode(), correctResponse, lang)




    # =============== TESTS ===============
    def test_line_is_on(self):
        statements = [
            ("en", "given line is on"),
            ("cs", "pokud je linka zapnutá"),
            ("cs", "pokud linka je zapnutá"),
        ]
        statementResponse = "{} lineon"

        for lang, statement in statements:
            self.singleStatementTest(lang, statement, statementResponse.format(g.get_kw(lang, "Given")))

    def test_station_xzx_free(self):
        statements = [
            ("en", "given station XZX is free"),
            ("cs", "pokud je stanice XZX volná"),
        ]
        statementResponse = "{} stationXZXFree"

        for lang, statement in statements:
            self.singleStatementTest(lang, statement, statementResponse.format(g.get_kw(lang,"Given")))

    def test_shuttle_y_goes_xzz(self):
        statements = [
            ("en", "given shuttle Y goes to station XZZ"),
            ("cs", "pokud vozíček Y jede do stanice XZZ"),
        ]
        statementResponse = "{} shuttleYDestination == XZZ"

        for lang, statement in statements:
            self.singleStatementTest(lang, statement, statementResponse.format(g.get_kw(lang,"Given")))

    def test_r1_picks_cube_one(self):
        statements = [
            ("en", "given robot R1 picks up cube 1"),
            ("en", "given robot R1 picks up cube one"),
            ("en", "given robot R1 picks up cube number 1"),
            ("cs", "pokud robot R1 zvedne kostku 1"),
            ("cs", "pokud robot R1 zvedne kostku jedna"),
            ("cs", "pokud robot R1 zvedne kostku číslo 1"),
        ]
        statementResponse = "{} robotR1ProgramNumber == 2"

        for lang, statement in statements:
            self.singleStatementTest(lang, statement, statementResponse.format(g.get_kw(lang,"Given")))

    def test_storage_is_empty(self):
        statements = [
            ("en", "given storage is empty"),
            ("cs", "pokud sklad je prázdný")
        ]
        statementResponse = "{} storageCount == 0"

        for lang, statement in statements:
            self.singleStatementTest(lang, statement, statementResponse.format(g.get_kw(lang,"Given")))

    @unittest.skip("TODO")
    def test_goes_to_station_at_robot(self):
        # TODO [issue #67] robot incorrectly marked as actor
        statements = [
            ("en", "given shuttle Z goes to station RRR by robot R"),
            ("cs", "pokud vozík Z jede do stanice RRR u robotu R")
        ]
        statementResponse = "{} shuttleZDestination == RRR"

        for lang, statement in statements:
            self.singleStatementTest(lang, statement, statementResponse.format(g.get_kw(lang,"Given")))


    def test_goes_to_station_at_table(self):
        statements = [
            ("en", "given shuttle Z goes to station RRR by table R"),
            ("cs", "pokud vozík Z jede do stanice RRR u stolu R")
        ]
        statementResponse = "{} shuttleZDestination == RRR"

        for lang, statement in statements:
            self.singleStatementTest(lang, statement, statementResponse.format(g.get_kw(lang,"Given")))




if __name__ == '__main__':
    unittest.TestCase.host = "localhost"
    unittest.TestCase.port = 5000
    unittest.main()



# -*- coding: utf-8 -*-
import difflib
import re
from gherkan.containers.signal_scenario import SignalScenario
from gherkan.containers.signal_batch import SignalBatch
from gherkan.utils.dictionaryMapper import DictionaryMapper
from gherkan.utils.dicts import sectionDict
import gherkan.utils.gherkin_keywords as g
import gherkan.utils.constants as c
import os
import logging

from gherkan.utils import logging_types


class Parser:
    def __init__(self):
        self.regex_flags = re.RegexFlag.IGNORECASE | re.RegexFlag.UNICODE
        self.sectionTypes = [g.GIVEN, g.WHEN, g.THEN]

        self.ROBOT_PROGRAMS_CZ = os.path.join(c.GHERKAN_ROOT_DIR, "utils", "RobotPrograms_cz.json")
        self.ROBOT_PROGRAMS_EN = os.path.join(c.GHERKAN_ROOT_DIR, "utils", "RobotPrograms_en.json")

        # active scenario for the current section
        self.scenario = None

        # path to file if the input is loaded from file
        self.filename = ""
        self.language = None
        self.negate_signal = False

    # def findSimilar(self, word, possibilities, cutoff=0.6):
    #     """
    #     Finds similar words. For future use: can be used to correct typos.
    #     """
    #     result = []
    #     s = difflib.SequenceMatcher()
    #     s.set_seq2(word)
    #     for idx, x in enumerate(possibilities):
    #         s.set_seq1(x)
    #         if s.real_quick_ratio() >= cutoff and s.quick_ratio() >= cutoff and s.ratio() >= cutoff:
    #             result.append((s.ratio(), x, idx))
    #
    #     return [idx for score, x, idx in result]

    def generateDirectiveExtractor(self, directiveName: str):
        # Generates a regural expression for directive extraction
        return re.compile(r"#\s*{}\W+(?P<result>\w+)".format(directiveName), flags=self.regex_flags)

    def getTextAfterColon(self, text: str):
        # Returns regex match with text after colon or None if regex fails
        reTextAfterColon = re.compile(r".*:\s*(?P<result>.+)", flags=self.regex_flags)

        return reTextAfterColon.search(text)

    def getTextAfterKeyword(self, text: str):
        # Returns regex match with text after whitespace or None if regex fails
        reTextAfterWhitespace = re.compile(r"^\s*\w+\s+(?P<result>.+)", flags=self.regex_flags)

        return reTextAfterWhitespace.search(text)

    def parseLanguage(self, line: str):
        # Parses directive on the current line
        # Directive == a "#" followed by a word followed by a colon and another word, e.g "# language: cs"

        # TODO not necessary, language is the only directive
        reDirective = self.generateDirectiveExtractor("language")
        directive = reDirective.search(line) or ""

        if not directive:
            logging.warning("Missing language in the input file, using 'en'",
                            extra={"type": logging_types.W_GENERAL_ERROR, "phrase" : line})
            return "en"

        return directive.group("result")

    def parseFeature(self, nlBatch: SignalBatch, lineNumber: int, textlines: list):
        # Parses the Feature section
        line = textlines[lineNumber]
        nextLine = textlines[lineNumber + 1]

        name = self.getTextAfterColon(line)

        if name:
            nlBatch.name = name.group("result")

        nlBatch.desc = nextLine

    def parseBackground(self, nlBatch: SignalBatch, lineNumber: int, textlines: list):
        # Parses the Background section
        nextLine = textlines[lineNumber + 1]
        cond = self.getTextAfterKeyword(nextLine)

        if cond:
            nlBatch.context = self.parseStatement(cond.group("result"), self.negate_signal)
        else:
            raise Exception("Context parsing failed!")

    def finalizeScenario(self, nlBatch: SignalBatch):
        # Finalizes the section with scenario
        self.scenario.contextStatements.append(nlBatch.context)
        nlBatch.addScenario(self.scenario)

    def parseScenario(self, nlBatch: SignalBatch, lineNumber: int, textlines: list):
        # Parses the Scenario section
        if self.scenario:
            # finalize the previous scenario
            self.finalizeScenario(nlBatch)

        # start a new scenario
        self.scenario = SignalScenario()

        # extract the scenario name
        line = textlines[lineNumber]
        scenarioName = self.getTextAfterColon(line)

        if scenarioName:
            self.scenario.name = scenarioName.group("result")
        else:
            logging.warning(
                "Missing name for a scenario on line {}, using random name", lineNumber,
                extra={"type": logging_types.W_GENERAL_ERROR, "phrase" : line})

    def parseScenarioSection(self, lineNumber: int, textlines: list, section: str):
        # Parses the Given / When / Then section inside the scenario section
        line = textlines[lineNumber]
        statement = self.getTextAfterKeyword(line)
        tree = self.parseStatement(statement.group("result"), self.negate_signal)

        if section == g.GIVEN:
            self.scenario.givenStatements.append(tree)
        elif section == g.WHEN:
            self.scenario.whenStatements.append(tree)
        elif section == g.THEN:
            self.scenario.thenStatements.append(tree)
        else:
            logging.warning(
                "Unknown section '{}' encountered on line {}", section, lineNumber,
                extra={"type" : logging_types.W_GENERAL_ERROR, "phrase" : textlines[lineNumber]})

    def getSectionList(self, textlines: list):
        # dictionary with section keywords
        sectionDM = DictionaryMapper(sectionDict)
        sectionList = []

        for i, line in enumerate(textlines):
            words = re.findall(r"^\s*(\w+)", line)
            words = sectionDM.filter(words)

            # TODO rewrite filter to return only section names
            if words:
                sectionList.append((i, words[0][1]))

        return sectionList



    def parse(self, textlines: list):
        """
        Parses the given text

        Parameters
        ----------
        textlines : list
            The textual data to be parsed

        signalBatch: SignalBatch
            SignalBatch container to be filled with data and returned.

        Returns
        -------
        SignalBatch:
            The SignalBatch container object.

        """
        nlBatch = SignalBatch(filename=self.filename)

        # extract language directive
        nlBatch.language = self.parseLanguage(textlines[0])
        self.language = nlBatch.language

        # retrieve the list of section with line numbers
        sectionList = self.getSectionList(textlines)

        textlines = self.mergeAndSections(textlines, sectionList)

        # Iterate over section denominators and process text accordingly
        for lineNumber, section in sectionList:
            if section == g.FEATURE:
                self.parseFeature(nlBatch, lineNumber, textlines)

            elif section == g.BACKGROUND:
                self.parseBackground(nlBatch, lineNumber, textlines)

            elif section == g.SCENARIO:
                self.parseScenario(nlBatch, lineNumber, textlines)

            elif section in self.sectionTypes and self.scenario is not None:
                self.parseScenarioSection(lineNumber, textlines, section)

        if self.scenario:
            self.finalizeScenario(nlBatch)

        return nlBatch

    def parseFile(self, inputFilePath: str):
        """
        Parameters
        ----------
        inputFilePath: str
            The full path, including filename, to the file containing the scenario batch
        """
        self.filename = os.path.basename(inputFilePath)

        with open(inputFilePath, "rt", encoding="utf-8") as file:
            lines = file.readlines()
            # remove empty lines and extra leading/trailing white space
            lines = [l.strip() for l in lines if len(l.strip()) > 0]

        return self.parse(lines)

    # virtual method, implemented in subclasses
    def parseStatement(self, statement: str, negate: bool):
        raise NotImplementedError()

    # virtual method, implemented in subclasses
    def mergeAndSections(self, textlines: list, sectionList: list):
        raise NotImplementedError()



import requests
import unittest

import gherkan.utils.gherkin_keywords as g
import gherkan.utils.constants as c

class TestSignalStatements(unittest.TestCase):

    def setUp(self):
        self.url = f"http://{self.__class__.host}:{self.__class__.port}/{{}}"

        resetResponse = requests.get(self.url.format("reset"))
        self.assertDictContainsSubset({"OK": True}, resetResponse.json())


    def compare_background_content(self, content, correct_background, lang):
        content_list = content.split("\n")
        content_list = [s.strip() for s in content_list]
        background_line_text = (g.get_kw(lang, "Background")) + ":"
        res_idx = content_list.index(background_line_text) + 1

        # case insensitive
        self.assertEqual(content_list[res_idx].lower(), correct_background.lower())


    def createStatementRequest(self, lang, statement):
        self.signalFile = f"""
            # language: {lang}

            Feature: StatementTest
              StatementTest dummy description
            Background:
              {statement}

            Scenario: given line is on
              Given lineOn
            """
        request = {'language': lang, 'scenarios': self.signalFile}

        return request

    def singleStatementTest(self, lang, statement, correctResponse):
        request = self.createStatementRequest(lang, statement)

        response = requests.post(self.url.format("signals"), data=request)
        try:
            response = response.json()
        except Exception as exception:
            self.fail(
                "An exception was raised while trying to convert response to JSON. "
                "This is most likely due to the fact that the response was of different type.\n"
                "The exception: {}\nResponse text (if possible):\n{}".format(
                    exception, response.text))
        else:
            self.assertDictContainsSubset({"OK": True}, response)

            response = requests.get(self.url.format("nltext"))

            self.compare_background_content(response.content.decode(), correctResponse, lang)




    # =============== TESTS ===============
    def test_line_is_on(self):
        statement = "given lineOn"
        statementResponses = [
            ("en", "Given line is on"),
            ("cs", "Pokud linka jest zapnutý")
        ]

        for lang, statementResponse in statementResponses:
            self.singleStatementTest(lang, statement, statementResponse)


    def test_station_xzx_free(self):
        statement = "given stationXZXFree"
        statementResponses = [
            ("en", "given station XZX is free"),
            ("cs", "pokud stanice XZX jest volný"),
        ]
        for lang, statementResponse in statementResponses:
            self.singleStatementTest(lang, statement, statementResponse)


    def test_shuttle_y_goes_zz(self):
        statement = "given shuttleYDestination == XZZ"
        statementResponses = [
            ("en", "given shuttle Y goes to station XZZ"),
            ("cs", "pokud vozíček Y pojede do stanice XZZ"),
        ]

        for lang, statementResponse in statementResponses:
            self.singleStatementTest(lang, statement, statementResponse)


    def test_r1_picks_cube_one(self):
        statement = "given robotR1ProgramNumber == 2"
        statementResponses = [
            ("en", "given robot R1 picks up cube 1"),
            ("cs", "pokud robot R1 zvedne kostka 1")
        ]

        for lang, statementResponse in statementResponses:
            self.singleStatementTest(lang, statement, statementResponse)


    def test_storage_is_empty(self):
        statement = "given storageCount == 0"
        statementResponses = [
            ("en", "given storage contain 0 pieces"),
            ("cs", "pokud sklad obsahuje 0 kusů")
        ]

        for lang, statementResponse in statementResponses:
            self.singleStatementTest(lang, statement, statementResponse)



if __name__ == '__main__':
    unittest.TestCase.host = "localhost"
    unittest.TestCase.port = 5000
    unittest.main()



import os, json, string
import en_coref_md
from pattern.en import parsetree, conjugate, lemma
import re
import gherkan.utils.constants as c
import gherkan.utils.gherkin_keywords as g
from gherkan.containers.NLPModelManager import NLPModelManager
from gherkan.encoder.NL2Temp import NLTempMatcher
from gherkan.utils.word_morpho import Finetuner
from gherkan.utils import logging_types
import logging


class RawNLParser:
    def __init__(self, nmm : NLPModelManager = None, feature=None, language="en", splitter="scenario"):
        """
        This class takes the raw text input and outputs the filled gherkin template in NL. It also updates
        the RobotPrograms.json with detected actions of selected subject
        :param feature: the actor we are talking about in the scenarios, i.e. robot R3
                (is filled when there is only "robot" without specification)
        :param language: language of the NL text (en/cs)
        :param subj_of_interest: subject of whom we collect actions and save them to RobotPrograms.json
        :param splitter: word which separates scenarios in NL text
        """
        self.lang = language
        self.split_word = splitter
        self.subj_of_interest = c.SUBJ_OF_INTEREST
        self.actor_uni = feature
        self.cleaner = NLTempMatcher(lang=self.lang, nmm=nmm, feature=feature)
        self.finetune = Finetuner(self.lang, nmm)
        self.text_raw = None
        self.robot_actions = []
        self.text_lines_normalized = []
        self.sentences = []
        self.keywords = [g.GIVEN, g.WHEN, g.THEN]
        if self.lang == "cs":
            self.keywords = [g.GIVEN_CS, g.WHEN_CS, g.THEN_CS]

    def parse(self, text_raw: str, background=False):
        """The main parser outputting the Given. When, Then template"""
        self.text_raw = text_raw
        scenarios = self.separate_scenarios(text_raw, self.split_word)
        for scenario in scenarios:
            logging.info("Processing scenario: {}", scenario)
            scenario = self.strip_extra_spaces(scenario)
            if not background:
              self.text_lines_normalized.append("Scenario: {}".format(scenario))
            text = self.replace_synonyms(scenario)
            self.find_patterns(text)
            phrases = self.prune_keywords(self.sentences)
            for phrase in phrases:
                    phrase = self.substitute_feature(phrase)
                    self.collect_subject_actions(phrase)
            for sentence in self.sentences:
                clean = self.strip_extra_spaces(sentence)
                if not background:
                   self.text_lines_normalized.append(clean)
                else:
                    return clean
            self.sentences = []

        logging.info("Detected actions connected with {}s: {}", self.subj_of_interest, self.robot_actions)

    def prune_keywords(self, text):
        """Separate sentences with AND/OR and also leave out keywords"""
        separators = {"en":["OR", "AND"], "cs":["NEBO", "A"]}
        pruned = []

        for phrase in text:
            sent = re.sub(r'\b({}|{}|{})\b'.format(*self.keywords), "", phrase, re.IGNORECASE)
            for x in separators[self.lang]:
                if isinstance(sent, list):
                    for y in sent:
                        sent_y = list(filter(None, re.split(r'\b{}\b'.format(x), y)))
                        sent_y = [x for x in sent_y if x.lower() != []]
                        pruned.extend(sent_y)
                else:
                    sent = list(filter(None, re.split(x, sent)))
                    sent = [x for x in sent if x.lower() != []]
        return pruned

    def get_text_lines(self):
        return self.text_lines_normalized

    def substitute_feature(self, phrase):
        """If there is no specification of the subject, we take it from the feature (e.g. robot > robot R1)"""
        actor = re.findall(r'{}\s+([A-Za-z0-9]{})\b'.format(self.subj_of_interest,"{1,3}"), phrase, re.IGNORECASE)
        if not actor:
            if self.subj_of_interest in self.actor_uni.lower():
               try:
                    actor_uni = " ".join([self.actor_uni.split()[0].lower(), self.actor_uni.split()[1].upper()])
               except:
                    actor_uni = self.actor_uni
            else:
                return phrase
            actor_new = actor_uni
            phrase = re.sub(r"\b{}\b".format(self.subj_of_interest), actor_new, phrase, re.IGNORECASE)
        return phrase

    def fix_word_order(self, text):
        """In Czech, change word order if verb is before subject"""
        phrase = re.sub(r'\b{}|{}\s'.format("Vynuť", "Uvolni"), "", text, re.IGNORECASE)
        w_order = ""
        new_v_pos = None
        for word in phrase.split():
            if word.isupper():
                w_order += "C"
            else:
                word_info = self.finetune.get_txt_info(word)
                w_order += (word_info[0]["pos"])
        if w_order[1] == "V":   # if verb is after keyword
            for idx, letter in enumerate(w_order[2:]):
                if letter != "N" and letter != "C":   # find first word which is not noun or number
                    if new_v_pos is None:
                        new_v_pos = idx + 2
                        if phrase is not text:
                            phrase = text
                            new_v_pos = new_v_pos + 1
                        new_phrase = phrase.split()
                        new_phrase.insert(new_v_pos, phrase.split()[1])
                        del new_phrase[1]
                        phrase = " ".join(new_phrase)
        else:
            phrase = text
        return phrase

    def replace_synonyms(self, command):
        """place words in command with synonyms defined in synonym_file"""
        synonyms_filename = os.path.join(
            c.GHERKAN_ROOT_DIR, "utils", "synonyms.json")
        with open(synonyms_filename, encoding="utf-8") as f:
            synonyms = json.load(f)
        for key in synonyms[self.lang]:
            for phrase in synonyms[self.lang][key]:
                if phrase in command.lower():
                    src_str = re.compile(r'\b{}\b'.format(phrase), re.IGNORECASE)
                    command = src_str.sub(key, command)
        command = self.strip_extra_spaces(re.sub(r"\stodelete", "", command))
        number_words = re.findall(r'\b({}|{})\b\s*[0-9]'.format("číslo", "number"), command, re.IGNORECASE)
        for x in number_words:
            command = re.sub(r'\b{}\b'.format(x), "", command, re.IGNORECASE)
        return command

    def find_patterns(self, text):
        """ Detects sentences framed with template keywords. Sentences without any keyword are detected as Then.
        CONDITION: if there is no "then", there must be at least a comma to separate the sentence"""
        for keyword in self.keywords:
            text = re.sub(r'\b{}\b'.format(keyword.lower()), keyword, text, re.MULTILINE)
        split = re.findall(
            r'\b(?:{0}|{1}|{2})\b.+?(?={0}|{1}|{2}|$)'.format(*self.keywords), text, re.IGNORECASE | re.MULTILINE)
        rest = None
        if len(split) >= 1:
           rest = list(filter(None, re.split(r'\.|,', split[-1], re.IGNORECASE)))
        try:
            if rest:
                for idx, x in enumerate(rest):
                    keyword = re.findall(r'\b{0}|{1}|{2}\b'.format(*self.keywords), x, re.IGNORECASE)
                    if not keyword:
                        rest[idx] = " ".join([self.keywords[2], x])
                split = split[:-1] + rest
        except:
            pass
        temp_dic = {"en": [["AND", "OR", "BUT"], ["and", "or", "but"]],
                    "cs": [["A", "NEBO", "ALE"], ["a", "nebo", "ale"]]}
        for x in split:
            phrase = x.translate(str.maketrans('', '', string.punctuation))
            phrase = self.strip_extra_spaces(phrase)
            for idx, keyword in enumerate(temp_dic[self.lang][0]):
                for x in range(1,4):
                    phrase = re.sub(r'\b{}\b'.format(temp_dic[self.lang][1][idx]), "".join([" ", keyword, " "]), phrase,re.IGNORECASE) # for some reason it sometimes misses some words on first try
            if self.lang == "cs":
                phrase = self.fix_word_order(phrase)
            self.sentences.append(phrase)
        self.sentences = list(set(self.sentences))
        corr_order = [[], [], []]  # Reorder list to Given,When,Then
        for idx, x in enumerate(self.keywords):
            for i in self.sentences:
                if x.lower() in i.lower():
                    i = re.sub(x.lower(), x, i)
                    corr_order[idx] = i
        self.sentences = list(filter(None, corr_order))
        return self.sentences

    def check_if_unique(self, program_dict, actor, action_lemma):
        default_verbs = ["začít", "skončit", "zastavit", "start", "stop", "end"]
        if actor in program_dict.keys():
            is_unique = True
            # find if the action is already in list
            for key in program_dict[actor]:
                if self.strip_extra_spaces(action_lemma) == program_dict[actor][key]:
                    is_unique = False
            if is_unique:
                if "program" in action_lemma.split()[-1] or action_lemma.split()[-1] in default_verbs:
                  pass
                else:
                    new_key = len(program_dict[actor]) + 1
                    program_dict[actor].update(
                        {new_key: self.strip_extra_spaces(action_lemma.lower())})
        else:
            if "program" in action_lemma.split()[-1] or action_lemma.split()[-1] in default_verbs:
                pass
            program_dict[actor] = {1: self.strip_extra_spaces(action_lemma)}
        return program_dict

    def generate_program_dict(self, make_new=True):
        """ make_new - creates a new dict and replaces old one. If False, programs are appended to current dict """
        program_dict = {}
        dict_path = os.path.join(c.GHERKAN_ROOT_DIR, "utils", "".join(["RobotPrograms_", self.lang, ".json"]))
        if not make_new:
            with open(dict_path, 'r', encoding="utf-8") as f:
                program_dict = json.load(f)
                f.close()
        if self.robot_actions:
            if self.lang == "en":
                for action in self.robot_actions:
                    actor_new = re.findall(r'({}\s+([a-z]+)*([0-9])*([a-z]+)*)'.format(self.subj_of_interest), action, re.IGNORECASE)
                    try:
                        actor = actor_new[0][0]
                    except:
                        pass
                    if not actor:  # if no specification, use the one defined in Feature
                        actor = self.actor_uni.lower()
                        action = re.sub("{}".format(self.subj_of_interest), "", action, re.IGNORECASE)
                    action_lemma = self.finetune.lemmatize_sentence(self.lang, "".join(action.lower().rsplit(actor.lower())))
                    program_dict = self.check_if_unique(program_dict, actor.lower(), action_lemma)
            elif self.lang == "cs":
                for action in self.robot_actions:
                    actor = re.findall(r'({}\s+(([a-z]+)*([0-9])*|([0-9])*([a-z]+)*))\s'.format(self.subj_of_interest), action, re.IGNORECASE)
                    if not actor:
                        actor = self.subj_of_interest
                        try:
                           actor_uni = " ".join([self.actor_uni.split()[0].lower(), self.actor_uni.split()[1].upper()])
                        except:
                            actor_uni = self.actor_uni
                        actor_new = actor_uni
                    else:
                        actor = actor[0][0]
                        actor_new = actor
                    action = self.cleaner.lemmatize(action)
                    action_lemma = re.sub(r'\b{}\b'.format(actor), "", action, re.IGNORECASE)
                    action_lemma = self.finetune.lemmatize_sentence(self.lang, action_lemma)
                    program_dict = self.check_if_unique(program_dict, actor_new.lower(), action_lemma)
            with open(dict_path, 'w', encoding="utf-8") as f:
                json.dump(program_dict, f, ensure_ascii=False, indent=4)
            logging.info("Updated program list: {}", program_dict)

    def strip_extra_spaces(self, text):
        stripped_spaces = re.sub(' +', ' ', text)
        stripped_text = stripped_spaces.strip()
        return stripped_text

    def collect_subject_actions(self, text):
        """ Detect any actions performed by subject of interest"""
        if self.lang == "en":
            tree = parsetree(text, tokenize=True, tags=True,
                             chunks=True, relations=True)
            for x in tree:  # now we select sentences with robot subjects or passive form
                for chunk in x.chunks:
                    if x.subjects:
                        if chunk.string in x.subjects[0].string:
                            if self.subj_of_interest.lower() in chunk.string.lower():  # active verb form
                                bool = self.cleaner.get_negative(x.string)
                                force, unforce = self.cleaner.get_force_unforce(x.string)
                                state, phrase = self.cleaner.get_state(x.string)
                                sentence_new = self.cleaner.get_cleanphr(phrase, bool, force, unforce)
                                self.robot_actions.append(sentence_new)
                if " ".join(["by", self.subj_of_interest]) in x.string.lower():  # handle passive verb form
                    verbs, nouns = ([] for i in range(2))
                    verb_object = (x.string).split(" by ")[0]
                    subject = (x.string).split(" by ")[1]
                    for chunk in x.chunks:
                        if chunk.type == "VP":
                            for word in chunk.words:
                                verbs.append(
                                    (lemma(word.string)))
                            if "be" in verbs:
                                verbs.remove("be")
                            rest = re.sub(r"\b{}\b".format(chunk.string), "", verb_object, re.IGNORECASE)
                    sent = subject + " " + ' '.join(verbs) + " " + rest
                    # sent = sent.translate(
                    #     str.maketrans('', '', string.punctuation))
                    bool = self.cleaner.get_negative(sent)
                    force, unforce = self.cleaner.get_force_unforce(sent)
                    sentence_new = self.cleaner.get_cleanphr(sent, bool, force, unforce)
                    self.robot_actions.append(sentence_new)

        elif self.lang == "cs":
            if self.subj_of_interest in text.split():
                bool = self.cleaner.get_negative(text)
                force, unforce = self.cleaner.get_force_unforce(text)
                state, phrase = self.cleaner.get_state(text)
                sentence_new = self.cleaner.get_cleanphr(phrase, bool, force, unforce)
                verb = self.finetune.find_verb(self.lang, sentence_new)
                if verb:
                    verb_lemma = self.finetune.lemmatize(self.lang, verb[0])
                    sentence_new = re.sub(verb[0], verb_lemma, sentence_new)
                self.robot_actions.append(sentence_new)
        # now remove duplicit actions:
        self.robot_actions = list(set(self.robot_actions))

    def solve_corefs(self, text):
        print("Loading Neuralcoref module...")
        nlc = en_coref_md.load()
        dok = nlc(text)
        if dok._.has_coref == 1:
            replaced_corefs = dok._.coref_resolved
        else:
            replaced_corefs = text
        print("Done")
        # print("Corref: " + replaced_corefs)

        return replaced_corefs

    def separate_scenarios(self, text, split_word):
        """Returns a list of scenarios separated by self.split_word. Upper case does not matter"""
        splitter = re.compile(r'\s*({})\s*'.format(split_word), re.IGNORECASE)
        split = re.split(splitter, text)
        split = list(filter(None, split))
        split_cl = [x for x in split if x.lower() != split_word]
        return split_cl

    def replace(self, string, substitutions):
        substrings = sorted(substitutions, key=len, reverse=True)
        regex = re.compile('|'.join(map(re.escape, substrings)))

        return regex.sub(lambda match: substitutions[match.group(0)], string)



import requests
import unittest
import re


class TestSignals(unittest.TestCase):

    def setUp(self):
        self.url = f"http://{self.__class__.host}:{self.__class__.port}/{{}}"

        resetResponse = requests.get(self.url.format("reset"))
        self.assertDictContainsSubset({"OK": True}, resetResponse.json())

        self.signalFile = """
            # language: en
            
            Feature: Montrac
              Montrac
            Background:
              Given lineOn
            
            Scenario: given line is on
              Given lineOn
            """

    def compare_multiline(self, testee, teststr):
        if not isinstance(testee, str) or not isinstance(teststr, str):
            self.assertFalse("No strings given to compare!")
            return

        self.assertEqual(testee.lower().replace("\n", " ").split(),
                         teststr.lower().replace("\n", " ").split())

    # @unittest.skip("TODO: skipped for debugging")
    def test_post_signal(self):
        data = {'language': 'cs', 'scenarios': self.signalFile}
        response = requests.post(self.url.format("signals"), data=data)
        response = response.json()
        self.assertDictContainsSubset({"OK": True}, response)

    # @unittest.skip("TODO: skipped for debugging")
    def test_post_and_get_nl(self):
        correctResponse = """
            # language: en
            
            Feature: Montrac
              Montrac
            Background:
              Given line is on
            
            Scenario: given line is on
              Given line is on
            """

        # wsr = re.compile(u"(^\s+)|(\s+$)", flags=re.MULTILINE | re.UNICODE)
        data = {'language': 'en', 'scenarios': self.signalFile}
        postResponse = requests.post(self.url.format("signals"), data=data)
        try:
            postResponse = postResponse.json()
        except Exception as exception:
            self.fail("An exception was raised whilel trying to convert response to JSON. This is most likely due to the fact that the response was of different type.\nThe exception: {}\nResponse text (if possible):\n{}".format(exception, postResponse.text))
        else:
            self.assertDictContainsSubset({"OK": True}, postResponse)
            response = requests.get(self.url.format("nltext"))

            self.compare_multiline(response.content.decode(), correctResponse)



if __name__ == '__main__':
    unittest.TestCase.host = "localhost"
    unittest.TestCase.port = 5000
    unittest.main()



import warnings

from gherkan.containers.NLPModelManager import NLPModelManager
from gherkan.decoder.SignalParser import SignalParser
from gherkan.encoder.NLFileWriter import NLFileWriter


def signal_to_nl(input_file_path: str, nmm: NLPModelManager):
    """
    Processes a signals file and saves it as NL file.

    Parameters
    ----------
    input_file_path path to signals file

    """

    # TODO find a better way to do this
    if input_file_path.endswith("_signals.feature"):
        output_file = input_file_path.replace("_signals.feature", ".feature")
    else:
        warnings.warn("File does not end with '_signals.feature', saving the NL file as {}.plain".format(input_file_path))
        output_file = input_file_path + ".plain"

    # parse the signals file
    sp = SignalParser()
    signalBatch = sp.parseFile(input_file_path)

    # save the NL file
    nlFileWriter = NLFileWriter(nmm)
    nlFileWriter.encode(signalBatch)
    nlFileWriter.write(output_file)


import requests
import unittest

import json
import os

import gherkan.utils.constants as c

class TestNLScenarios(unittest.TestCase):

    def setUp(self):
        self.url = f"http://{self.__class__.host}:{self.__class__.port}/{{}}"
        self.test_path = os.path.join(c.GHERKAN_ROOT_DIR, "tests", "scenarios")

        resetResponse = requests.get(self.url.format("reset"))
        self.assertDictContainsSubset({"OK": True}, resetResponse.json())


    def compare_multiline(self, testee, teststr):
        if not isinstance(testee, str) or not isinstance(teststr, str):
            self.assertFalse("No strings given to compare!")
            return

        self.assertEqual(testee.lower().replace("\n", " ").split(),
                         teststr.lower().replace("\n", " ").split())


    def scenarioTest(self, input_filename, correct_output_filename):
        with open(os.path.join(self.test_path, input_filename), "rt", encoding="utf-8") as f:
            request = json.load(f)

        with open(os.path.join(self.test_path, correct_output_filename), "rt", encoding="utf-8") as f:
            correctResponse = f.read()

        response = requests.post(self.url.format("nltext"), data=request)
        try:
            response = response.json()
        except Exception as exception:
            self.fail(
                "An exception was raised while trying to convert response to JSON. "
                "This is most likely due to the fact that the response was of different type.\n"
                "The exception: {}\nResponse text (if possible):\n{}".format(
                    exception, response.text))
        else:
            # self.assertDictContainsSubset({"errors": False}, response)
            response = requests.get(self.url.format("signals"))

            content = response.content.decode()

            # print(content)
            self.compare_multiline(content, correctResponse)


    def get_files(self, scenario_name, level, lang):
        nl_filename = f"{scenario_name}_{level}_NL_{lang}.json"
        signal_filename = f"{scenario_name}_{level}_signals_{lang}.txt"

        return nl_filename, signal_filename



    # =============== TESTS ===============

    def test_montrac_easy_en(self):
        nl_filename, signal_filename = self.get_files("Montrac", "easy", "en")
        self.scenarioTest(nl_filename, signal_filename)

    def test_montrac_easy_cs(self):
        nl_filename, signal_filename = self.get_files("Montrac", "easy", "cs")
        self.scenarioTest(nl_filename, signal_filename)

    def test_bgscenarioDif_basic_en(self):
        nl_filename, signal_filename = self.get_files("BgScenarioDiff", "basic", "en")
        self.scenarioTest(nl_filename, signal_filename)

    def test_bgscenarioDif_easy_en(self):
        nl_filename, signal_filename = self.get_files("BgScenarioDiff", "easy", "en")
        self.scenarioTest(nl_filename, signal_filename)

    def test_r1FillStorage_easy_en(self):
        nl_filename, signal_filename = self.get_files("r1FillStorage", "easy", "en")
        self.scenarioTest(nl_filename, signal_filename)

    def test_r1FillStorage_easy_cs(self):
        nl_filename, signal_filename = self.get_files("r1FillStorage", "easy", "cs")
        self.scenarioTest(nl_filename, signal_filename)

    def test_r1NormalSeq_easy_en(self):
        nl_filename, signal_filename = self.get_files("r1NormalSeq", "easy", "en")
        self.scenarioTest(nl_filename, signal_filename)

    @unittest.skip("TODO")
    def test_r1NormalSeq_easy_cs(self):
        nl_filename, signal_filename = self.get_files("r1NormalSeq", "easy", "cs")
        self.scenarioTest(nl_filename, signal_filename)

    def test_r2FillStorage_easy_en(self):
        nl_filename, signal_filename = self.get_files("r2FillStorage", "easy", "en")
        self.scenarioTest(nl_filename, signal_filename)

    def test_r2FillStorage_easy_cs(self):
        nl_filename, signal_filename = self.get_files("r2FillStorage", "easy", "cs")
        self.scenarioTest(nl_filename, signal_filename)

    def test_r3FillStorage_easy_en(self):
        nl_filename, signal_filename = self.get_files("r3FillStorage", "easy", "en")
        self.scenarioTest(nl_filename, signal_filename)

    def test_r3FillStorage_easy_cs(self):
        nl_filename, signal_filename = self.get_files("r3FillStorage", "easy", "cs")
        self.scenarioTest(nl_filename, signal_filename)


if __name__ == '__main__':
    unittest.TestCase.host = "localhost"
    unittest.TestCase.port = 5000
    unittest.main()



from .tree import Tree
from .visitors import Transformer, Visitor, v_args, Discard
from .visitors import InlineTransformer, inline_args   # XXX Deprecated
from .exceptions import ParseError, LexError, GrammarError, UnexpectedToken, UnexpectedInput, UnexpectedCharacters
from .lexer import Token
from .lark import Lark

__version__ = "0.6.6"




import os
import gherkan.utils.constants as c
from gherkan.containers.NLPModelManager import NLPModelManager
from gherkan.flask_api.signal_to_nl import signal_to_nl

# Path to the text file with the signal batch

dir = os.path.join(c.DATA_DIR, 'output')
input_file = os.path.join(dir, 'case_SignalParser_signals.feature')

nmm = NLPModelManager()
nmm.load_all_models()

# saves automatically to file without the "_signals" suffix
signal_to_nl(input_file, nmm)

output_file = input_file.replace("_signals", "")

with open(output_file, "rt", encoding="utf-8") as f:
    text = f.read()
    print(text)




from gherkan.containers.NLPModelManager import NLPModelManager
from gherkan.decoder.SignalParser import SignalParser

#Translates variables in tree to NLP or back to signals based on the tranType
# (in each node of the tree a variable data.VariableNLP will be filled)
#   st - original statementTree
from gherkan.processing.TreeProcessor import TreeProcessor

if __name__ == "__main__":
    sp = SignalParser()
    st = sp.parseStatement(
     "robotR2ProgramNumber == 1 && robotR1ProgramNumber == 2 && part2Destination == 0 && BoxFull == 1")

    nmm = NLPModelManager()
    nmm.load_all_models()

    tp = TreeProcessor(language="en", nmm=nmm, feature="robot R2")

    # loads yaml file with templates
    tp.load_templ_dic('utils/templates_dic.yaml')
    st.root = tp.process_tree(st.root, dir=TreeProcessor.Direction.SIGNAL_TO_NL)

    # result is saved in node.data.variableNL_full
    print(st)

    # TODO linearize tree


import re


# === X<preposition>Y ===
def prepositionRE(preposition: str):
    """
    detects constructs in the form of "X<preposition>Y"
    In regex, (?P=<X>) denotes a capturing group with the name "X". I.e. it can be easily retrieved (see example below).
    """
    return re.compile(r"(?P<X>\w+)" + preposition + r"(?P<Y>\w+)", re.IGNORECASE)


# Create regex for <X>On<Y>
reXOnY = prepositionRE("on")  # generate regex to detect "On" cases. Use any other proposition to generate other regexes.
match = reXOnY.search("part2OnTable")  # parse the string using the regex
print(match.groups())  # print all captured groups
print("Found that {} is located on {}.".format(match.group("X"), match.group("Y")))  # Demonstration of the retrieval of the capturing groups by name


# === RobotXProgramNumber
robotXProgram = re.compile(r"robot(?P<robotName>\w+)programnumber", re.IGNORECASE)
match = robotXProgram.search("robotN3ProgramNumber")  # parse the string using the regex
print("Robot name was '{}'".format(match.group("robotName")))



# -*- coding: utf-8 -*-
from gherkan.containers.StatementTree import StatementTree
from gherkan.decoder.Parser import Parser

from lark import Lark
from gherkan.utils import gherkin_keywords as g

class SignalParser(Parser):
    def __init__(self):
        super().__init__()

        self.statement_grammar = """
            ?start: expression

            ?expression: statement
                | "(" expression ")"
                | expression "&&" expression       -> and
                | expression "||" expression       -> or
                | "!" "(" expression ")"           -> not

            ?statement: function
                | NAME "==" atom                -> equality
                | NAME "!=" atom                -> inequality
                | NAME                          -> bool
                | "!" statement                 -> not

            ?function: function_name "(" NAME "," atom ")"

            ?function_name: "edge"           -> edge 
                          | "force"          -> force
                          | "unforce"        -> unforce
            
            ?atom: NAME
                | NUMBER

            %import common.CNAME -> NAME
            %import common.NUMBER
            %import common.WS_INLINE

            %ignore WS_INLINE
        """


    def parseStatement(self, statement: str, negate=False):
        parser = Lark(self.statement_grammar, parser='earley', ambiguity='resolve', propagate_positions=True)
        tree = parser.parse(statement)

        # print(tree.pretty())

        st = StatementTree(statement)
        st.buildFromSignalTree(tree, negate)

        # print(st)

        return st

    def mergeAndSections(self, textlines: list, sectionList: list):
        """
        Naïve approach for parsing "And" sections. Find the lines with "And" sections and merge their statements
        with the previous line by "&&" operator.
        """
        for lineNumber, section in sectionList:
            if section == g.AND:
                statement = self.getTextAfterKeyword(textlines[lineNumber]).group("result")
                textlines[lineNumber - 1] += " && {}".format(statement)
                textlines[lineNumber] = ""

        return textlines

    def setNegateSignal(self):
        self.negate_signal = True


import sys
import os

from gherkan.utils import constants as c


from ufal.morphodita import *
import logging

class CZMorph:
    def __init__(self, nlp_model_cs):
        self.tagger = nlp_model_cs
        self.tokenizer = self.tagger.newTokenizer()

        if self.tokenizer is None:
            logging.error("No tokenizer is defined for the supplied model!")

    def get_pos_info(self, text):
        analysed = []
        forms = Forms()
        lemmas = TaggedLemmas()
        tokens = TokenRanges()
        self.tokenizer.setText(text)
        while self.tokenizer.nextSentence(forms, tokens):
          # Tag - explanations at http://ufal.mff.cuni.cz/pdt2.0/doc/manuals/en/m-layer/html/ch02s02s01.html
          self.tagger.tag(forms, lemmas)
          for i in range(len(lemmas)):
            lemma = lemmas[i]
            token = tokens[i]
            pos = self.encode_entities(lemma.tag[0])
            pos_det = self.encode_entities(lemma.tag[1])
            gender = self.encode_entities(lemma.tag[2])
            num = self.encode_entities(lemma.tag[3])
            case = self.encode_entities(lemma.tag[4])
            # poss_gender = self.encode_entities(lemma.tag[5])
            # poss_num = self.encode_entities(lemma.tag[6])
            person = self.encode_entities(lemma.tag[7])
            # tense = self.encode_entities(lemma.tag[8])
            negation = self.encode_entities(lemma.tag[10])

            analysed.append({"word":self.encode_entities(text[token.start : token.start + token.length]),"pos": pos,
                             "pos_det":pos_det, "case":case, "num":num, "person":person, "gender":gender, "negate":negation})
        return analysed

    def encode_entities(self,text):
      return text.replace('&', '&amp;').replace('<', '&lt;').replace('>', '&gt;').replace('"', '&quot;')


import requests
import unittest

import gherkan.utils.gherkin_keywords as g
import gherkan.utils.constants as c



class TestNLStatements(unittest.TestCase):

    def setUp(self):
        self.url = f"http://{self.__class__.host}:{self.__class__.port}/{{}}"

        resetResponse = requests.get(self.url.format("reset"))
        self.assertDictContainsSubset({"OK": True}, resetResponse.json())


    def compare_scenario_content(self, content, correct_scenario, lang):
        content_list = content.split("\n")
        content_list = [s.strip() for s in content_list]

        idx = 0

        for i, line in enumerate(content_list):
            if line.startswith(g.get_kw(lang, "Scenario")):
                idx = i+1
                break

        scenario = " ".join(content_list[idx:])

        self.assertEqual(scenario.lower().split(),
                         correct_scenario.lower().replace("\n", " ").split())


    def createStatementRequest(self, lang, statement):
        request = {
            "feature": "StatementTest",
            "feature_desc": "StatementTest dummy description",
            "language": lang,
            # dummy background, only scenario is tested
            "background": "given line is on" if lang == "en"
                            else "pokud linka je zapnutá",
            "text_raw": "scenario " + statement
        }
        return request

    def singleStatementTest(self, lang, statement, correctResponse):
        request = self.createStatementRequest(lang, statement)

        response = requests.post(self.url.format("nltext"), data=request)
        try:
            response = response.json()
        except Exception as exception:
            self.fail(
                "An exception was raised while trying to convert response to JSON. "
                "This is most likely due to the fact that the response was of different type.\n"
                "The exception: {}\nResponse text (if possible):\n{}".format(
                    exception, response.text))
        else:
            self.assertDictContainsSubset({"errors": False}, response)

            response = requests.get(self.url.format("signals"))

            self.compare_scenario_content(response.content.decode(), correctResponse, lang)




    # =============== TESTS ===============
    @unittest.skip("TODO")
    def test_a31_fill_words(self):
        statements = [
            ("cs", "Ehm pokud je ten no ehm vozíček X na takzvané jak bych to řekl pozici Y"),
        ]
        statementResponse = "{} shuttleXAtStationY == 1"

        for lang, statement in statements:
            self.singleStatementTest(lang, statement, statementResponse.format(g.get_kw(lang, "Given")))

    @unittest.skip("TODO")
    def test_a32_synonyms_homonyms(self):
        statements = [
            ("cs", "Jestliže je vozíček X v pozici Y, tak robot R1 vyloží všechny kostičky"),
            ("cs", "Pokud stojí vozíček X ve stanici Y, pak robot R1 vyndá všechny kostičky."),
            ("cs", "Pokud se nachází vozíček X v místě Y, potom robot vyskládá všechny kostičky."),
        ]
        statementResponse = "{} shuttleXAtStationY {} robotR1ProgramNumber == 6"

        for lang, statement in statements:
            self.singleStatementTest(lang, statement, statementResponse.format(g.get_kw(lang, "Given"),
                                                                               g.get_kw(lang, "Then")))

    @unittest.skip("TODO")
    def test_a33_common_subject(self):
        statements = [
            ("cs", "Když součástky 1, 2 a 3 jsou na stole"),
        ]
        statementResponse = "{} part1OnTable && part2OnTable && part3OnTable {} lineOn"

        for lang, statement in statements:
            self.singleStatementTest(lang, statement, statementResponse.format(g.get_kw(lang, "When")))

    @unittest.skip("TODO")
    def test_a34_multiple_actions(self):
        statements = [
            ("cs", "Když robot R3 zvedne kostičku a vyloží kostičku."),
        ]
        statementResponse = "{} (robotR3ProgramNumber == 3) && (robotR3ProgramNumber == 4)"

        for lang, statement in statements:
            self.singleStatementTest(lang, statement, statementResponse.format(g.get_kw(lang, "When")))

    @unittest.skip("TODO")
    def test_a35_action_type(self):
        statements = [
            ("cs", "Pokud je linka zapnutá, pak vynuť vozíček X jede do stanice Y"),
            ("en", "Given line is on, then force shuttle X goes to station Y."),
        ]
        statementResponse = "{} lineOn {} force(shuttleXDestination, y)"

        for lang, statement in statements:
            self.singleStatementTest(lang, statement, statementResponse.format(g.get_kw(lang, "Given"),
                                                                               g.get_kw(lang, "Then")))

    @unittest.skip("TODO")
    def test_a36_unexpressed_subject(self):
        statements = [
            ("cs", "Když je vozíček X ve stanici Y, pak jede do stanice Z."),
        ]
        statementResponse = "{} shuttleXAtStationY {} shuttleXDestination == Z"

        for lang, statement in statements:
            self.singleStatementTest(lang, statement, statementResponse.format(g.get_kw(lang, "Given"),
                                                                               g.get_kw(lang, "Then")))


    def test_a37_multilingual(self):
        statements = [
            ("en", "When shuttle XX goes to station Z."),
            ("cs", "Když vozíček XX jede do stanice Z."),
        ]
        statementResponse = "{} shuttleXXDestination == Z"

        for lang, statement in statements:
            self.singleStatementTest(lang, statement, statementResponse.format(g.get_kw(lang, "When")))



if __name__ == '__main__':
    unittest.TestCase.host = "localhost"
    unittest.TestCase.port = 5000
    unittest.main()




import warnings

from gherkan.containers.NLPModelManager import NLPModelManager
from gherkan.decoder.SignalParser import SignalParser
from gherkan.encoder.SignalFileWriter import SignalFileWriter


def signal_to_negated_signal(input_file_path: str, nmm: NLPModelManager):
    """
    Processes a signals file and saves it as NL file.

    Parameters
    ----------
    input_file_path path to signals file

    """

    # TODO find a better way to do this
    if input_file_path.endswith("_signals.feature"):
        output_file = input_file_path.replace("_signals.feature", "_neg_signals.feature")
    else:
        output_file = input_file_path + "_neg_signals.feature"

    # parse the signals file
    sp = SignalParser()
    sp.setNegateSignal()
    signalBatch = sp.parseFile(input_file_path)

    # save the NL file
    signalFileWriter = SignalFileWriter(nmm)
    signalFileWriter.encode(signalBatch, no_processing=True)
    signalFileWriter.write(output_file)

    return output_file


class SignalScenario():
    """
    Container class for test scenario encoded into signal.

    ...

    Attributes
    ----------

    name : str
        name of the scenario

    contextStatements : list
        context for the "given" conditions

    givenStatements : list
        list Given conditions

    whenStatements : list
        list When conditions

    thenStatements : list
        list Then conditions

    """

    def __init__(self, name=""):
        self.name = name
        self.contextStatements = []
        self.givenStatements = []
        self.whenStatements = []
        self.thenStatements = []



from flask import jsonify, request
from flask_restful import Resource, reqparse
from .. import API_FSA

class NLScenario(Resource):

    def get(self):
        """
        response = {
                "language":"en<OR>cz",
                "background": "background text",
                "description": "optional description",
                "scenarios": [
                    "scenario1 NL text",
                    "scenario2 NL text"
                ]
        }
        """
        pass

